"""Labware state store tests.

DEPRECATED: Testing LabwareStore independently of LabwareView is no
longer helpful. Try to add new tests to test_labware_state.py, where they can be
tested together, treating LabwareState as a private implementation detail.
"""

from datetime import datetime
from typing import Optional

import pytest

from opentrons_shared_data.deck.types import DeckDefinitionV5
from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition,
    LabwareDefinition2,
    Parameters2,
)

from .command_fixtures import (
    create_comment_command,
)
from opentrons.calibration_storage.helpers import uri_from_details
from opentrons.protocol_engine.actions import (
    AddLabwareDefinitionAction,
    AddLabwareOffsetAction,
    SucceedCommandAction,
)
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.labware import LabwareState, LabwareStore
from opentrons.protocol_engine.types import (
    OFF_DECK_LOCATION,
    DeckSlotLocation,
    LabwareOffset,
    LabwareOffsetCreateInternal,
    LabwareOffsetVector,
    LegacyLabwareOffsetLocation,
    LoadedLabware,
    OnAddressableAreaOffsetLocationSequenceComponent,
)
from opentrons.types import DeckSlotName


@pytest.fixture
def subject(
    ot2_standard_deck_def: DeckDefinitionV5,
) -> LabwareStore:
    """Get a LabwareStore test subject."""
    return LabwareStore(
        deck_definition=ot2_standard_deck_def,
        deck_fixed_labware=[],
    )


def test_initial_state(
    ot2_standard_deck_def: DeckDefinitionV5,
    subject: LabwareStore,
) -> None:
    """It should create the labware store with preloaded fixed labware."""
    assert subject.state == LabwareState(
        deck_definition=ot2_standard_deck_def,
        labware_by_id={},
        labware_offsets_by_id={},
        definitions_by_uri={},
    )


def test_handles_add_labware_offset(
    subject: LabwareStore,
) -> None:
    """It should add the labware offset to the state and add the ID."""
    request = LabwareOffsetCreateInternal(
        definitionUri="offset-definition-uri",
        legacyLocation=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_1),
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1")
        ],
        vector=LabwareOffsetVector(x=1, y=2, z=3),
    )

    resolved_offset = LabwareOffset(
        id="offset-id",
        createdAt=datetime(year=2021, month=1, day=2),
        definitionUri="offset-definition-uri",
        location=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_1),
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1")
        ],
        vector=LabwareOffsetVector(x=1, y=2, z=3),
    )

    subject.handle_action(
        AddLabwareOffsetAction(
            labware_offset_id="offset-id",
            created_at=datetime(year=2021, month=1, day=2),
            request=request,
        )
    )

    assert subject.state.labware_offsets_by_id == {"offset-id": resolved_offset}


@pytest.mark.parametrize(
    "display_name, offset_id", [("display-name", "offset-id"), (None, None)]
)
def test_handles_load_labware(
    subject: LabwareStore,
    well_plate_def: LabwareDefinition,
    display_name: Optional[str],
    offset_id: Optional[str],
) -> None:
    """It should add the labware data to the state."""
    offset_request = LabwareOffsetCreateInternal(
        definitionUri="offset-definition-uri",
        legacyLocation=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_1),
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1")
        ],
        vector=LabwareOffsetVector(x=1, y=2, z=3),
    )

    command = create_comment_command()

    expected_definition_uri = uri_from_details(
        load_name=well_plate_def.parameters.loadName,
        namespace=well_plate_def.namespace,
        version=well_plate_def.version,
    )

    expected_labware_data = LoadedLabware(
        id="test-labware-id",
        loadName=well_plate_def.parameters.loadName,
        definitionUri=expected_definition_uri,
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        offsetId=offset_id,
        displayName=display_name,
    )

    subject.handle_action(
        AddLabwareOffsetAction(
            request=offset_request,
            labware_offset_id="offset-id",
            created_at=datetime(year=2021, month=1, day=2),
        )
    )
    subject.handle_action(
        SucceedCommandAction(
            command=command,
            state_update=update_types.StateUpdate(
                loaded_labware=update_types.LoadedLabwareUpdate(
                    labware_id="test-labware-id",
                    definition=well_plate_def,
                    offset_id=offset_id,
                    display_name=display_name,
                    new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
                ),
            ),
        )
    )

    assert subject.state.labware_by_id["test-labware-id"] == expected_labware_data

    assert subject.state.definitions_by_uri[expected_definition_uri] == well_plate_def


def test_handles_reload_labware(
    subject: LabwareStore,
    well_plate_def: LabwareDefinition,
) -> None:
    """It should override labware data in the state."""
    command = create_comment_command()

    subject.handle_action(
        SucceedCommandAction(
            command=command,
            state_update=update_types.StateUpdate(
                loaded_labware=update_types.LoadedLabwareUpdate(
                    labware_id="test-labware-id",
                    definition=well_plate_def,
                    offset_id=None,
                    display_name="display-name",
                    new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
                ),
            ),
        )
    )
    expected_definition_uri = uri_from_details(
        load_name=well_plate_def.parameters.loadName,
        namespace=well_plate_def.namespace,
        version=well_plate_def.version,
    )
    assert (
        subject.state.labware_by_id["test-labware-id"].definitionUri
        == expected_definition_uri
    )

    offset_request = LabwareOffsetCreateInternal(
        definitionUri="offset-definition-uri",
        legacyLocation=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_1),
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1")
        ],
        vector=LabwareOffsetVector(x=1, y=2, z=3),
    )
    subject.handle_action(
        AddLabwareOffsetAction(
            request=offset_request,
            labware_offset_id="offset-id",
            created_at=datetime(year=2021, month=1, day=2),
        )
    )
    comment_command_2 = create_comment_command(
        command_id="comment-id-1",
    )
    subject.handle_action(
        SucceedCommandAction(
            command=comment_command_2,
            state_update=update_types.StateUpdate(
                labware_location=update_types.LabwareLocationUpdate(
                    new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
                    offset_id="offset-id",
                    labware_id="test-labware-id",
                )
            ),
        )
    )

    expected_labware_data = LoadedLabware(
        id="test-labware-id",
        loadName=well_plate_def.parameters.loadName,
        definitionUri=expected_definition_uri,
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        offsetId="offset-id",
        displayName="display-name",
    )
    assert subject.state.labware_by_id["test-labware-id"] == expected_labware_data
    assert subject.state.definitions_by_uri[expected_definition_uri] == well_plate_def


def test_handles_add_labware_definition(
    subject: LabwareStore,
    well_plate_def: LabwareDefinition,
) -> None:
    """It should add the labware definition to the state."""
    expected_uri = uri_from_details(
        load_name=well_plate_def.parameters.loadName,
        namespace=well_plate_def.namespace,
        version=well_plate_def.version,
    )

    subject.handle_action(AddLabwareDefinitionAction(definition=well_plate_def))

    assert subject.state.definitions_by_uri[expected_uri] == well_plate_def


def test_handles_move_labware(
    subject: LabwareStore,
    well_plate_def: LabwareDefinition,
) -> None:
    """It should update labware state with new location & offset."""
    comment_command = create_comment_command()
    offset_request = LabwareOffsetCreateInternal(
        definitionUri="offset-definition-uri",
        legacyLocation=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_1),
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1")
        ],
        vector=LabwareOffsetVector(x=1, y=2, z=3),
    )
    subject.handle_action(
        AddLabwareOffsetAction(
            request=offset_request,
            labware_offset_id="old-offset-id",
            created_at=datetime(year=2021, month=1, day=2),
        )
    )
    subject.handle_action(
        SucceedCommandAction(
            command=comment_command,
            state_update=update_types.StateUpdate(
                loaded_labware=update_types.LoadedLabwareUpdate(
                    labware_id="my-labware-id",
                    definition=well_plate_def,
                    offset_id=None,
                    display_name="display-name",
                    new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
                ),
            ),
        )
    )

    comment_2 = create_comment_command(
        command_id="my-command-id",
    )
    subject.handle_action(
        SucceedCommandAction(
            command=comment_2,
            state_update=update_types.StateUpdate(
                labware_location=update_types.LabwareLocationUpdate(
                    labware_id="my-labware-id",
                    offset_id="my-new-offset",
                    new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
                ),
            ),
        )
    )

    assert subject.state.labware_by_id["my-labware-id"].location == DeckSlotLocation(
        slotName=DeckSlotName.SLOT_1
    )
    assert subject.state.labware_by_id["my-labware-id"].offsetId == "my-new-offset"


def test_handles_move_labware_off_deck(
    subject: LabwareStore,
    well_plate_def: LabwareDefinition,
) -> None:
    """It should update labware state with new location & offset."""
    comment_command = create_comment_command()
    offset_request = LabwareOffsetCreateInternal(
        definitionUri="offset-definition-uri",
        legacyLocation=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_1),
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1")
        ],
        vector=LabwareOffsetVector(x=1, y=2, z=3),
    )
    subject.handle_action(
        AddLabwareOffsetAction(
            request=offset_request,
            labware_offset_id="old-offset-id",
            created_at=datetime(year=2021, month=1, day=2),
        )
    )
    subject.handle_action(
        SucceedCommandAction(
            command=comment_command,
            state_update=update_types.StateUpdate(
                loaded_labware=update_types.LoadedLabwareUpdate(
                    labware_id="my-labware-id",
                    definition=well_plate_def,
                    offset_id=None,
                    display_name="display-name",
                    new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
                ),
            ),
        )
    )

    comment_2 = create_comment_command(
        command_id="my-command-id",
    )
    subject.handle_action(
        SucceedCommandAction(
            command=comment_2,
            state_update=update_types.StateUpdate(
                labware_location=update_types.LabwareLocationUpdate(
                    labware_id="my-labware-id",
                    new_location=OFF_DECK_LOCATION,
                    offset_id=None,
                )
            ),
        )
    )
    assert subject.state.labware_by_id["my-labware-id"].location == OFF_DECK_LOCATION
    assert subject.state.labware_by_id["my-labware-id"].offsetId is None


def test_handle_batch_labware_loaded_update(
    subject: LabwareStore,
    well_plate_def: LabwareDefinition,
) -> None:
    """It should consume batch_loaded_labware updates."""
    request = LabwareOffsetCreateInternal(
        definitionUri="offset-definition-uri",
        legacyLocation=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_1),
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1")
        ],
        vector=LabwareOffsetVector(x=1, y=2, z=3),
    )
    subject.handle_action(
        AddLabwareOffsetAction(
            labware_offset_id="some-offset-id",
            created_at=datetime(year=2021, month=1, day=2),
            request=request,
        )
    )
    comment_command = create_comment_command()
    subject.handle_action(
        SucceedCommandAction(
            command=comment_command,
            state_update=update_types.StateUpdate(
                batch_loaded_labware=update_types.BatchLoadedLabwareUpdate(
                    new_locations_by_id={
                        "some-labware-id": DeckSlotLocation(
                            slotName=DeckSlotName.SLOT_1
                        ),
                        "some-other-labware-id": OFF_DECK_LOCATION,
                    },
                    offset_ids_by_id={
                        "some-labware-id": "some-offset-id",
                        "some-other-labware-id": None,
                    },
                    display_names_by_id={
                        "some-labware-id": "some-display-name",
                        "some-other-labware-id": None,
                    },
                    definitions_by_id={
                        "some-labware-id": LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                            schemaVersion=2,
                            namespace="namespace-1",
                            version=1,
                            parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                                loadName="load-name-1"
                            ),
                        ),
                        "some-other-labware-id": LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                            schemaVersion=2,
                            namespace="namespace-2",
                            version=2,
                            parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                                loadName="load-name-2"
                            ),
                        ),
                    },
                )
            ),
        )
    )
    some_labware = subject.state.labware_by_id["some-labware-id"]
    some_other_labware = subject.state.labware_by_id["some-other-labware-id"]
    assert some_labware.id == "some-labware-id"
    assert some_labware.location == DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
    assert some_labware.loadName == "load-name-1"
    assert some_labware.definitionUri == "namespace-1/load-name-1/1"
    assert some_labware.offsetId == "some-offset-id"
    assert some_labware.displayName == "some-display-name"

    assert some_other_labware.id == "some-other-labware-id"
    assert some_other_labware.location == OFF_DECK_LOCATION
    assert some_other_labware.loadName == "load-name-2"
    assert some_other_labware.definitionUri == "namespace-2/load-name-2/2"
    assert some_other_labware.offsetId is None
    assert some_other_labware.displayName is None


def test_handle_batch_labware_location_update(
    subject: LabwareStore,
    well_plate_def: LabwareDefinition,
) -> None:
    """It should consume batch_labware_location updates."""
    request = LabwareOffsetCreateInternal(
        definitionUri="offset-definition-uri",
        legacyLocation=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_1),
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1")
        ],
        vector=LabwareOffsetVector(x=1, y=2, z=3),
    )
    subject.handle_action(
        AddLabwareOffsetAction(
            labware_offset_id="some-offset-id",
            created_at=datetime(year=2021, month=1, day=2),
            request=request,
        )
    )
    comment_command = create_comment_command()
    subject.handle_action(
        SucceedCommandAction(
            command=comment_command,
            state_update=update_types.StateUpdate(
                batch_loaded_labware=update_types.BatchLoadedLabwareUpdate(
                    new_locations_by_id={
                        "some-labware-id": DeckSlotLocation(
                            slotName=DeckSlotName.SLOT_1
                        ),
                        "some-other-labware-id": OFF_DECK_LOCATION,
                    },
                    offset_ids_by_id={
                        "some-labware-id": "some-offset-id",
                        "some-other-labware-id": None,
                    },
                    display_names_by_id={
                        "some-labware-id": "some-display-name",
                        "some-other-labware-id": None,
                    },
                    definitions_by_id={
                        "some-labware-id": LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                            schemaVersion=2,
                            namespace="namespace-1",
                            version=1,
                            parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                                loadName="load-name-1"
                            ),
                        ),
                        "some-other-labware-id": LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                            schemaVersion=2,
                            namespace="namespace-2",
                            version=2,
                            parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                                loadName="load-name-2"
                            ),
                        ),
                    },
                )
            ),
        )
    )
    comment_command2 = create_comment_command()
    subject.handle_action(
        SucceedCommandAction(
            command=comment_command2,
            state_update=update_types.StateUpdate(
                batch_labware_location=update_types.BatchLabwareLocationUpdate(
                    new_locations_by_id={
                        "some-labware-id": OFF_DECK_LOCATION,
                        "some-other-labware-id": DeckSlotLocation(
                            slotName=DeckSlotName.SLOT_1
                        ),
                    },
                    new_offset_ids_by_id={
                        "some-labware-id": None,
                        "some-other-labware-id": "some-offset-id",
                    },
                )
            ),
        )
    )
    some_labware = subject.state.labware_by_id["some-labware-id"]
    some_other_labware = subject.state.labware_by_id["some-other-labware-id"]
    assert some_labware.location == OFF_DECK_LOCATION
    assert some_labware.offsetId is None

    assert some_other_labware.location == DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
    assert some_other_labware.offsetId == "some-offset-id"



from __future__ import annotations

import asyncio
import logging
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Literal,
    Mapping,
    Optional,
    cast,
)

from opentrons_shared_data.errors.exceptions import (
    FlexStackerHopperLabwareError,
    FlexStackerShuttleLabwareError,
    FlexStackerShuttleMissingError,
    FlexStackerShuttleNotEmptyError,
    FlexStackerStallError,
)
from opentrons_shared_data.module import load_tof_baseline_data

from opentrons.config import feature_flags as ff
from opentrons.drivers.flex_stacker.abstract import AbstractFlexStackerDriver
from opentrons.drivers.flex_stacker.driver import (
    FlexStackerDriver,
)
from opentrons.drivers.flex_stacker.simulator import SimulatingDriver
from opentrons.drivers.flex_stacker.types import (
    AxisParams,
    Direction,
    HardwareRevision,
    LEDColor,
    LEDPattern,
    MoveParams,
    MoveResult,
    StackerAxis,
    StallGuardParams,
    TOFDetection,
    TOFMeasurementResult,
    TOFSensor,
    TOFSensorMode,
    TOFSensorState,
    TOFSensorStatus,
)
from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.hardware_control.execution_manager import ExecutionManager
from opentrons.hardware_control.modules import mod_abc, update
from opentrons.hardware_control.modules.types import (
    FlexStackerData,
    FlexStackerStatus,
    HopperDoorState,
    LatchState,
    LiveData,
    ModuleDisconnectedCallback,
    ModuleErrorCallback,
    ModuleType,
    PlatformState,
    StackerAxisState,
    UploadFunction,
)
from opentrons.hardware_control.poller import Poller, Reader
from opentrons.hardware_control.types import StatusBarState, StatusBarUpdateEvent

log = logging.getLogger(__name__)

POLL_PERIOD = 2.0
SIMULATING_POLL_PERIOD = POLL_PERIOD / 20.0

DFU_PID = "df11"

# Maximum distance in mm the axis can travel.
MAX_TRAVEL = {
    StackerAxis.X: 194.0,
    StackerAxis.Z: 139.5,
    StackerAxis.L: 22.0,
}

# Min/Max height in mm of labware stack to store/dispense
MIN_LABWARE_HEIGHT = 4.0
MAX_LABWARE_HEIGHT = 102.5

# The offset in mm to subtract from MAX_TRAVEL when moving an axis before we home.
# This lets us use `move_axis` to move fast, leaving the axis OFFSET mm
# from the limit switch. Then we can use `home_axis` to move the axis the rest
# of the way until we trigger the expected limit switch.
HOME_OFFSET_SM = 5.0
HOME_OFFSET_MD = 10.0

# The labware platform will contact the labware this mm before the platform
# touches the +Z endstop.
PLATFORM_OFFSET = 2.25

# Should put the bottom of the plate above this mm above the latch when dispensing.
# Should put the bottom of the plate this mm below the latch when storing.
LATCH_CLEARANCE = 2.5

# TOF Baseline Configs
# These are generated manually with the help of the tof_analysis.py tool
# Which can be found in `hardware_testing/tools/tof-analysis/README.md` where
# it goes over using use the tool.
TOF_DETECTION_CONFIG = {
    TOFSensor.X: {
        Direction.EXTEND: TOFDetection(
            TOFSensor.X,
            zones=[5, 6, 7],
            bins=list(range(30, 40)),
            threshold=1000,
        ),
        Direction.RETRACT: TOFDetection(
            TOFSensor.X,
            zones=[5, 6, 7],
            bins=list(range(17, 30)),
            threshold=1000,
        ),
    },
    TOFSensor.Z: {
        Direction.EXTEND: TOFDetection(
            TOFSensor.Z,
            zones=[1, 2, 3],
            bins=list(range(15, 63)),
            threshold=1000,
        ),
        Direction.RETRACT: TOFDetection(
            TOFSensor.Z,
            zones=[1, 2, 3],
            bins=list(range(15, 63)),
            threshold=1000,
        ),
    },
}


# Stallguard defaults
STALLGUARD_CONFIG = {
    StackerAxis.X: StallGuardParams(StackerAxis.X, True, 0),
    StackerAxis.Z: StallGuardParams(StackerAxis.Z, True, 2),
}

# Motion Parameter defaults
STACKER_MOTION_CONFIG = {
    StackerAxis.X: {
        "home": AxisParams(
            run_current=1.5,  # Amps RMS
            hold_current=0.75,
            move_params=MoveParams(
                max_speed=10.0,  # mm/s
                acceleration=100.0,  # mm/s^2
                max_speed_discont=40.0,  # mm/s
            ),
        ),
        "move": AxisParams(
            run_current=1.2,
            hold_current=0.75,
            move_params=MoveParams(
                max_speed=200.0,
                acceleration=1500.0,
                max_speed_discont=40.0,
            ),
        ),
    },
    StackerAxis.Z: {
        "home": AxisParams(
            run_current=1.5,
            hold_current=1.5,
            move_params=MoveParams(
                max_speed=10.0,
                acceleration=100.0,
                max_speed_discont=25.0,
            ),
        ),
        "move": AxisParams(
            run_current=1.5,
            hold_current=1.5,
            move_params=MoveParams(
                max_speed=150.0,
                acceleration=500.0,
                max_speed_discont=25.0,
            ),
        ),
    },
    StackerAxis.L: {
        "home": AxisParams(
            run_current=1.2,
            hold_current=0.5,
            move_params=MoveParams(
                max_speed=100.0,
                acceleration=800.0,
                max_speed_discont=40.0,
            ),
        ),
        "move": AxisParams(
            run_current=1.2,
            hold_current=0.5,
            move_params=MoveParams(
                max_speed=100.0,
                acceleration=800.0,
                max_speed_discont=40.0,
            ),
        ),
    },
}


class FlexStacker(mod_abc.AbstractModule):
    """Hardware control interface for an attached Flex-Stacker module."""

    MODULE_TYPE = ModuleType.FLEX_STACKER

    @classmethod
    async def build(
        cls,
        port: str,
        usb_port: USBPort,
        hw_control_loop: asyncio.AbstractEventLoop,
        execution_manager: ExecutionManager,
        disconnected_callback: ModuleDisconnectedCallback,
        error_callback: ModuleErrorCallback,
        poll_interval_seconds: float | None = None,
        simulating: bool = False,
        sim_model: Optional[str] = None,
        sim_serial_number: Optional[str] = None,
    ) -> "FlexStacker":
        """
        Build a FlexStacker

        Args:
            port: The port to connect to
            usb_port: USB Port
            execution_manager: Execution manager.
            hw_control_loop: The event loop running in the hardware control thread.
            poll_interval_seconds: Poll interval override.
            simulating: whether to build a simulating driver
            loop: Loop
            sim_model: The model name used by simulator
            disconnected_callback: Callback to inform the module controller that the device was disconnected

        Returns:
            FlexStacker instance
        """
        driver: AbstractFlexStackerDriver
        if not simulating:
            driver = await FlexStackerDriver.create(port=port, loop=hw_control_loop)
            poll_interval_seconds = poll_interval_seconds or POLL_PERIOD
        else:
            driver = SimulatingDriver(serial_number=sim_serial_number)
            poll_interval_seconds = poll_interval_seconds or SIMULATING_POLL_PERIOD

        reader = FlexStackerReader(driver=driver)
        poller = Poller(reader=reader, interval=poll_interval_seconds)
        module = cls(
            port=port,
            usb_port=usb_port,
            driver=driver,
            reader=reader,
            poller=poller,
            device_info=(await driver.get_device_info()).to_dict(),
            hw_control_loop=hw_control_loop,
            execution_manager=execution_manager,
            disconnected_callback=disconnected_callback,
            error_callback=error_callback,
        )

        # Enable stallguard
        for axis, config in STALLGUARD_CONFIG.items():
            await driver.set_stallguard_threshold(
                axis, config.enabled, config.threshold
            )

        try:
            await poller.start()
        except Exception:
            log.exception(f"First read of Flex-Stacker on port {port} failed")

        return module

    def __init__(
        self,
        port: str,
        usb_port: USBPort,
        driver: AbstractFlexStackerDriver,
        reader: FlexStackerReader,
        poller: Poller,
        device_info: Mapping[str, str],
        hw_control_loop: asyncio.AbstractEventLoop,
        execution_manager: ExecutionManager,
        disconnected_callback: ModuleDisconnectedCallback,
        error_callback: ModuleErrorCallback,
    ):
        super().__init__(
            port=port,
            usb_port=usb_port,
            hw_control_loop=hw_control_loop,
            execution_manager=execution_manager,
            disconnected_callback=disconnected_callback,
            error_callback=error_callback,
        )
        self._device_info = device_info
        self._driver = driver
        self._reader = reader
        self._poller = poller
        self._stall_detected = False
        self._stacker_status = FlexStackerStatus.IDLE
        self._last_status_bar_event: Optional[StatusBarUpdateEvent] = None
        self._should_identify = False
        # Set initialized callback
        self._unsubscribe_init = reader.set_initialized_callback(
            self._initialized_callback
        )
        self._unsubscribe_error = reader.set_error_callback(self._async_error_callback)

    async def _initialized_callback(self) -> None:
        """Called by the reader once the module is initialized."""
        if self._last_status_bar_event:
            await self._handle_status_bar_event(self._last_status_bar_event)

    def _async_error_callback(self, exception: Exception) -> None:
        self.error_callback(exception)

    async def cleanup(self) -> None:
        """Stop the poller task"""
        await self._poller.stop()
        await self._driver.disconnect()

    @classmethod
    def name(cls) -> str:
        """Used for picking up serial port symlinks"""
        return "flexstacker"

    def firmware_prefix(self) -> str:
        """The prefix used for looking up firmware"""
        return "flex-stacker"

    @staticmethod
    def _model_from_revision(revision: Optional[str]) -> str:
        """Defines the revision -> model mapping"""
        return "flexStackerModuleV1"

    def model(self) -> str:
        return self._model_from_revision(self._device_info.get("model"))

    @property
    def latch_state(self) -> LatchState:
        """The state of the latch."""
        return LatchState.from_state(self.limit_switch_status[StackerAxis.L])

    @property
    def platform_state(self) -> PlatformState:
        """The state of the platform."""
        return self._reader.platform_state

    @property
    def initialized(self) -> bool:
        """The stacker is ready..."""
        return self._reader.initialized

    @property
    def hopper_door_state(self) -> HopperDoorState:
        """The status of the hopper door."""
        return HopperDoorState.from_state(self._reader.hopper_door_closed)

    @property
    def limit_switch_status(self) -> Dict[StackerAxis, StackerAxisState]:
        """The status of the Limit switches."""
        return self._reader.limit_switch_status

    @property
    def install_detected(self) -> bool:
        """Whether the stacker is installed on Flex."""
        return self._reader.installation_detected

    @property
    def device_info(self) -> Mapping[str, str]:
        return self._device_info

    @property
    def status(self) -> FlexStackerStatus:
        """Module status or error state details."""
        return self._stacker_status

    @property
    def is_simulated(self) -> bool:
        return isinstance(self._driver, SimulatingDriver)

    @property
    def live_data(self) -> LiveData:
        data: FlexStackerData = {
            "latchState": self.latch_state.value,
            "platformState": self.platform_state.value,
            "hopperDoorState": self.hopper_door_state.value,
            "installDetected": self.install_detected,
            "errorDetails": self._reader.error,
        }
        return {"status": self.status.value, "data": data}

    @property
    def should_identify(self) -> bool:
        return self._should_identify

    async def prep_for_update(self) -> str:
        await self._poller.stop()
        await self._driver.stop_motors()
        await self._driver.enter_programming_mode()
        # flex stacker has three unique "devices" over DFU
        dfu_info = await update.find_dfu_device(pid=DFU_PID, expected_device_count=3)
        return dfu_info

    def bootloader(self) -> UploadFunction:
        return update.upload_via_dfu

    async def deactivate(self, must_be_running: bool = True) -> None:
        await self._driver.stop_motors()

    async def set_led_state(
        self,
        power: float,
        color: Optional[LEDColor] = None,
        pattern: Optional[LEDPattern] = None,
        duration: Optional[int] = None,
        reps: Optional[int] = None,
    ) -> None:
        """Sets the statusbar state."""
        return await self._driver.set_led(
            power, color=color, pattern=pattern, duration=duration, reps=reps
        )

    async def move_axis(
        self,
        axis: StackerAxis,
        direction: Direction,
        distance: float,
        speed: Optional[float] = None,
        acceleration: Optional[float] = None,
        current: Optional[float] = None,
    ) -> bool:
        """Move the axis in a direction by the given distance in mm."""
        default = STACKER_MOTION_CONFIG[axis]["move"]
        old_run_current = self._reader.motion_params[axis].run_current
        new_run_current = current if current is not None else default.run_current
        if new_run_current != old_run_current:
            await self._driver.set_run_current(axis, new_run_current)
            self._reader.motion_params[axis].run_current = new_run_current

        old_hold_current = self._reader.motion_params[axis].hold_current
        new_hold_current = default.hold_current
        if new_hold_current != old_hold_current:
            await self._driver.set_ihold_current(axis, new_hold_current)
            self._reader.motion_params[axis].hold_current = new_hold_current

        motion_params = default.move_params.update(
            max_speed=speed, acceleration=acceleration
        )
        distance = direction.distance(distance)
        res = await self._driver.move_in_mm(axis, distance, params=motion_params)
        if res == MoveResult.STALL_ERROR:
            self._stall_detected = True
            raise FlexStackerStallError(self.device_info["serial"], axis)
        return res == MoveResult.NO_ERROR

    async def home_axis(
        self,
        axis: StackerAxis,
        direction: Direction,
        speed: Optional[float] = None,
        acceleration: Optional[float] = None,
        current: Optional[float] = None,
    ) -> bool:
        default = STACKER_MOTION_CONFIG[axis]["home"]
        old_run_current = self._reader.motion_params[axis].run_current
        new_run_current = current if current is not None else default.run_current
        if new_run_current != old_run_current:
            await self._driver.set_run_current(axis, new_run_current)
            self._reader.motion_params[axis].run_current = new_run_current

        old_hold_current = self._reader.motion_params[axis].hold_current
        new_hold_current = default.hold_current
        if new_hold_current != old_hold_current:
            await self._driver.set_ihold_current(axis, new_hold_current)
            self._reader.motion_params[axis].hold_current = new_hold_current

        motion_params = default.move_params.update(
            max_speed=speed, acceleration=acceleration
        )
        success = await self._driver.move_to_limit_switch(
            axis=axis, direction=direction, params=motion_params
        )
        await self._reader.get_limit_switch_status()
        if success == MoveResult.STALL_ERROR:
            self._stall_detected = True
            raise FlexStackerStallError(self.device_info["serial"], axis)
        return success == MoveResult.NO_ERROR

    async def close_latch(
        self,
        velocity: Optional[float] = None,
        acceleration: Optional[float] = None,
    ) -> bool:
        """Close the latch, dropping any labware its holding."""
        success = await self.home_axis(
            StackerAxis.L,
            Direction.RETRACT,
            speed=velocity,
            acceleration=acceleration,
        )
        # Check that the latch is closed.
        await self._reader.get_limit_switch_status()
        return (
            success
            and self.limit_switch_status[StackerAxis.L] == StackerAxisState.EXTENDED
        )

    async def open_latch(
        self,
        velocity: Optional[float] = None,
        acceleration: Optional[float] = None,
    ) -> bool:
        """Open the latch."""
        # The latch only has one limit switch, so we have to travel a fixed distance
        # to open the latch.
        success = await self.move_axis(
            StackerAxis.L,
            Direction.EXTEND,
            distance=MAX_TRAVEL[StackerAxis.L],
            speed=velocity,
            acceleration=acceleration,
        )
        # Check that the latch is opened.
        await self._reader.get_limit_switch_status()
        return (
            success
            and self.limit_switch_status[StackerAxis.L] == StackerAxisState.RETRACTED
        )

    async def dispense_labware(
        self,
        labware_height: float,
        enforce_hopper_lw_sensing: bool = True,
        enforce_shuttle_lw_sensing: bool = True,
    ) -> None:
        """Dispenses the next labware in the stacker."""
        self.verify_labware_height(labware_height)
        await self._prepare_for_action()

        if enforce_hopper_lw_sensing:
            # TODO: re-enable this function after TOF calibration is implemented.
            # Until then, we should also check the TOF X sensor before raising the error
            # await self.verify_hopper_labware_presence(Direction.EXTEND, True)
            hopper_empty = not await self.labware_detected(
                StackerAxis.Z, Direction.EXTEND
            )

        # Move platform along the X and make sure we DONT detect labware
        await self._move_and_home_axis(StackerAxis.X, Direction.RETRACT, HOME_OFFSET_MD)
        if enforce_shuttle_lw_sensing:
            await self.verify_shuttle_labware_presence(Direction.RETRACT, False)

        # Move platform along the Z axis
        await self._move_and_home_axis(StackerAxis.Z, Direction.EXTEND, HOME_OFFSET_SM)

        # Transfer
        await self.open_latch()
        latch_clear_distance = labware_height + PLATFORM_OFFSET - LATCH_CLEARANCE
        await self.move_axis(StackerAxis.Z, Direction.RETRACT, latch_clear_distance)
        await self.close_latch()

        # Move Z down the rest of the way
        z_distance = MAX_TRAVEL[StackerAxis.Z] - latch_clear_distance - HOME_OFFSET_SM
        await self.move_axis(StackerAxis.Z, Direction.RETRACT, z_distance)
        await self.home_axis(StackerAxis.Z, Direction.RETRACT)

        if enforce_shuttle_lw_sensing:
            try:
                await self.verify_shuttle_labware_presence(Direction.RETRACT, True)
            except FlexStackerShuttleLabwareError:
                # No labware detected on the shuttle, so we need to check what the Z TOF
                # sensor says about the hopper
                if hopper_empty:
                    # homing here so we don't have to modify the error recovery flow
                    await self._move_and_home_axis(
                        StackerAxis.X, Direction.EXTEND, HOME_OFFSET_MD
                    )
                    raise FlexStackerHopperLabwareError(
                        self.device_info["serial"],
                        labware_expected=True,
                    ) from None
                raise

        await self._move_and_home_axis(StackerAxis.X, Direction.EXTEND, HOME_OFFSET_MD)

    async def store_labware(
        self,
        labware_height: float,
        enforce_shuttle_lw_sensing: bool = True,
    ) -> None:
        """Stores a labware in the stacker."""
        self.verify_labware_height(labware_height)
        await self._prepare_for_action()

        # Move the X and check that labware is detected
        await self._move_and_home_axis(StackerAxis.X, Direction.RETRACT, HOME_OFFSET_MD)
        if enforce_shuttle_lw_sensing:
            await self.verify_shuttle_labware_presence(Direction.RETRACT, True)

        # Move the Z so the labware sits right under the labware already stored
        latch_clear_distance = labware_height + PLATFORM_OFFSET - LATCH_CLEARANCE
        distance = MAX_TRAVEL[StackerAxis.Z] - latch_clear_distance
        await self.move_axis(StackerAxis.Z, Direction.EXTEND, distance)

        await self.open_latch()
        # Move the labware the rest of the way at half move speed to increase torque.
        remaining_z = latch_clear_distance - HOME_OFFSET_SM
        speed_z = STACKER_MOTION_CONFIG[StackerAxis.Z]["move"].move_params.max_speed / 2
        await self.move_axis(StackerAxis.Z, Direction.EXTEND, remaining_z, speed_z)
        await self.home_axis(StackerAxis.Z, Direction.EXTEND, speed_z)
        await self.close_latch()

        # Move the Z down and check that labware is not detected.
        await self._move_and_home_axis(StackerAxis.Z, Direction.RETRACT, HOME_OFFSET_MD)
        if enforce_shuttle_lw_sensing:
            await self.verify_shuttle_labware_presence(Direction.RETRACT, False)

        # Move the X to the gripper position
        await self._move_and_home_axis(StackerAxis.X, Direction.EXTEND, HOME_OFFSET_MD)

    async def _move_and_home_axis(
        self,
        axis: StackerAxis,
        direction: Direction,
        offset: float = 0,
        speed: Optional[float] = None,
    ) -> bool:
        """Move the axis in a direction by the given offset in mm and home it.

        Warning: It is assumed that the axis is already in a known state
        before this function gets called. Do not use this function if the axis
        has not been homed/has recently stalled."""
        distance = MAX_TRAVEL[axis] - offset
        await self.move_axis(axis, direction, distance, speed)
        return await self.home_axis(axis, direction, speed)

    async def _prepare_for_action(self) -> None:
        """Helper to prepare axis for dispensing or storing labware."""
        # TODO: check if we need to home first
        await self.home_axis(StackerAxis.X, Direction.EXTEND)
        await self.home_axis(StackerAxis.Z, Direction.RETRACT)
        await self.close_latch()
        await self.verify_shuttle_location(PlatformState.EXTENDED)

    async def home_all(self, ignore_latch: bool = False) -> None:
        """Home all axes based on current state, assuming normal operation.

        If ignore_latch is True, we will not attempt to close the latch. This
        is useful when we want the shuttle to be out of the way for error
        recovery (e.g. when the latch is stuck open).
        """
        await self._reader.get_installation_detected()
        await self._reader.get_limit_switch_status()
        await self._reader.get_platform_sensor_state()

        # Z axis is unknown, lets move it up in case it is holding a labware
        if not ignore_latch:
            if self.limit_switch_status[StackerAxis.Z] == StackerAxisState.UNKNOWN:
                if self.latch_state == LatchState.OPENED:
                    # let's make sure the latch is opened all the way before homging the Z
                    await self.open_latch()
                    # self.latch_state is OPENED, so we need to home Z in the EXTEND direction
                    await self.home_axis(StackerAxis.Z, Direction.EXTEND)
            await self.close_latch()

        if (
            # if the platform is on the z or if x has not been homed
            self.platform_state == PlatformState.UNKNOWN
            or self.limit_switch_status[StackerAxis.X] == StackerAxisState.UNKNOWN
        ):
            # if the z is not retracted, we need to make sure the x is retracted
            # so we can retract the z properly later
            if self.limit_switch_status[StackerAxis.Z] != StackerAxisState.RETRACTED:
                await self.home_axis(StackerAxis.X, Direction.RETRACT)
            else:
                await self.home_axis(StackerAxis.X, Direction.EXTEND)

        # Finally, retract Z and extend X if they are not already
        await self.home_axis(StackerAxis.Z, Direction.RETRACT)
        await self.home_axis(StackerAxis.X, Direction.EXTEND)

    async def labware_detected(
        self,
        axis: StackerAxis,
        direction: Direction,
        histogram: Optional[TOFMeasurementResult] = None,
        baseline: Optional[Dict[int, List[float]]] = None,
    ) -> bool:
        """Detect labware on the TOF sensor using the `baseline` method

        NOTE: This method is still under development and is inconsistent when detecting
        labware on the X axis in the Extended position. We can consistently detect
        labware on the Z, but we need to do more data collection and testing
        to validate this method.
        """
        dir_str = cast(Literal["extend", "retract"], str(direction))
        sensor = TOFSensor.X if axis == StackerAxis.X else TOFSensor.Z
        baseline = (
            baseline or load_tof_baseline_data(self.model())[sensor.value][dir_str]
        )
        config = TOF_DETECTION_CONFIG[sensor][direction]

        # Take a histogram reading and determine if labware was detected
        histogram = histogram or await self._driver.get_tof_histogram(sensor)
        for zone in config.zones:
            raw_data = histogram.bins[zone]
            baseline_data = baseline[zone]
            for bin in config.bins:
                # We need to ignore raw photon count below N photons as
                # it becomes inconsistent to detect labware given false positives.
                if raw_data[bin] < config.threshold:
                    continue
                delta = raw_data[bin] - baseline_data[bin]
                if delta > 0:
                    return True
        return False

    async def verify_shuttle_location(self, expected: PlatformState) -> None:
        """Verify the shuttle is present and in the expected location."""
        await self._reader.get_platform_sensor_state()
        # Validate the platform state matches, ignore EXTENDED checks on EVT
        if self.platform_state != expected:
            if (
                self.device_info["model"] == HardwareRevision.EVT.value
                and expected == PlatformState.EXTENDED
            ):
                return
            else:
                raise FlexStackerShuttleMissingError(
                    self.device_info["serial"], expected, self.platform_state
                )

    async def verify_shuttle_labware_presence(
        self, direction: Direction, labware_expected: bool
    ) -> None:
        """Check whether or not a labware is detected on the shuttle."""
        if ff.flex_stacker_tof_sensors_disabled():
            return
        result = await self.labware_detected(StackerAxis.X, direction)
        if labware_expected != result:
            if labware_expected:
                raise FlexStackerShuttleLabwareError(
                    self.device_info["serial"],
                    shuttle_state=self.platform_state,
                    labware_expected=labware_expected,
                )
            raise FlexStackerShuttleNotEmptyError(
                self.device_info["serial"],
                shuttle_state=self.platform_state,
                labware_expected=labware_expected,
            )

    async def verify_hopper_labware_presence(
        self, direction: Direction, labware_expected: bool
    ) -> None:
        """Check whether or not a labware is detected inside the hopper."""
        if ff.flex_stacker_tof_sensors_disabled():
            return
        result = await self.labware_detected(StackerAxis.Z, direction)
        if labware_expected != result:
            raise FlexStackerHopperLabwareError(
                self.device_info["serial"],
                labware_expected=labware_expected,
            )

    def verify_labware_height(self, labware_height: float) -> None:
        """Check that the labware height is within valid range."""
        if labware_height < MIN_LABWARE_HEIGHT or labware_height > MAX_LABWARE_HEIGHT:
            raise ValueError(
                f"Labware height must be between {MIN_LABWARE_HEIGHT}-{MAX_LABWARE_HEIGHT}mm."
                "Received {labware_height}mm."
            )

    def event_listener(self, event: Any) -> None:
        if isinstance(event, StatusBarUpdateEvent):
            self._last_status_bar_event = event
            asyncio.run_coroutine_threadsafe(
                self._handle_status_bar_event(event), self._loop
            )

    async def _handle_status_bar_event(self, event: StatusBarUpdateEvent) -> None:  # noqa: C901
        if event.enabled and self.initialized:
            match event.state:
                case StatusBarState.RUNNING:
                    await self.set_led_state(0.5, LEDColor.GREEN, LEDPattern.STATIC)
                case StatusBarState.PAUSED:
                    if (
                        self.hopper_door_state == HopperDoorState.OPENED
                        or self.should_identify
                    ):
                        await self._stacker_bar_pause()
                    else:
                        await self._stacker_bar_idle()
                case StatusBarState.IDLE:
                    if self.hopper_door_state == HopperDoorState.OPENED:
                        await self._stacker_bar_pause()
                    else:
                        await self._stacker_bar_idle()
                case StatusBarState.HARDWARE_ERROR:
                    if self.should_identify:
                        await self.set_led_state(
                            0.5, LEDColor.RED, LEDPattern.FLASH, duration=300
                        )
                    else:
                        await self._stacker_bar_idle()
                case StatusBarState.SOFTWARE_ERROR:
                    await self.set_led_state(0.5, LEDColor.YELLOW, LEDPattern.STATIC)
                case StatusBarState.ERROR_RECOVERY:
                    if self.hopper_door_state == HopperDoorState.OPENED:
                        await self._stacker_bar_pause()
                    elif self.should_identify:
                        await self.set_led_state(
                            0.5, LEDColor.YELLOW, LEDPattern.PULSE, duration=2000
                        )
                    else:
                        await self._stacker_bar_idle()
                case StatusBarState.RUN_COMPLETED:
                    await self.set_led_state(0.5, LEDColor.GREEN, LEDPattern.PULSE)
                case StatusBarState.UPDATING:
                    await self.set_led_state(0.5, LEDColor.WHITE, LEDPattern.PULSE)
                case _:
                    await self._stacker_bar_idle()

    async def _stacker_bar_pause(self) -> None:
        await self.set_led_state(0.5, LEDColor.BLUE, LEDPattern.PULSE, duration=2000)

    async def _stacker_bar_idle(self) -> None:
        await self.set_led_state(0.5, LEDColor.WHITE, LEDPattern.STATIC)

    async def identify(self, start: bool, color_name: Optional[str] = None) -> None:
        """Identify the module."""
        reps = -1 if start else 0
        color = LEDColor.from_name(color_name or LEDColor.BLUE.name)
        await self.set_led_state(0.5, color, LEDPattern.PULSE, reps=reps)
        if not start and self._last_status_bar_event:
            await self._handle_status_bar_event(self._last_status_bar_event)

    def set_stacker_identify(self, state: bool) -> None:
        self._should_identify = state

    def cleanup_persistent(self) -> None:
        """Reset persistent data on the module that should not exist outside of a run."""
        self.set_stacker_identify(False)


class FlexStackerReader(Reader):
    error: Optional[str]

    def __init__(self, driver: AbstractFlexStackerDriver) -> None:
        self.error: Optional[str] = None
        self._driver = driver
        self.limit_switch_status = {
            axis: StackerAxisState.UNKNOWN for axis in StackerAxis
        }
        self.tof_sensor_status: Dict[TOFSensor, TOFSensorStatus] = {
            s: TOFSensorStatus(
                s, TOFSensorState.INITIALIZING, TOFSensorMode.UNKNOWN, False
            )
            for s in TOFSensor
        }
        self.motion_params: Dict[StackerAxis, AxisParams] = {
            axis: AxisParams(0, 0, MoveParams(0, 0, 0)) for axis in StackerAxis
        }
        self.platform_state = PlatformState.UNKNOWN
        self.hopper_door_closed = False
        self.initialized = False
        self.installation_detected = False
        self._refresh_state = False
        self._initialized_callback: Optional[Callable[[], Awaitable[None]]] = None
        self._error_callback: Optional[Callable[[Exception], None]] = None

    def set_initialized_callback(
        self, callback: Callable[[], Awaitable[None]]
    ) -> Callable[[], None]:
        """Sets the callback used when done initializing the module."""
        self._initialized_callback = callback
        return self._remove_init_callback

    def _remove_init_callback(self) -> None:
        self._initialized_callback = None

    def set_error_callback(
        self, error_callback: Callable[[Exception], None]
    ) -> Callable[[], None]:
        """Register a handler for asynchronous hardware errors."""
        self._error_callback = error_callback
        return self._remove_error_callback

    def _remove_error_callback(self) -> None:
        self._error_callback = None

    async def read(self) -> None:
        await self.get_door_closed()
        await self.get_platform_sensor_state()
        if not self.initialized or self._refresh_state:
            initialized = True
            await self.get_installation_detected()
            await self.get_limit_switch_status()
            await self.get_motion_parameters()
            for sensor, status in self.tof_sensor_status.items():
                if status.state == TOFSensorState.INITIALIZING:
                    status = await self._driver.get_tof_sensor_status(sensor)
                    self.tof_sensor_status[sensor] = status
                    initialized &= status.ok

            self._refresh_state = False
            # We are done initializing, sync the led state
            if not self.initialized and initialized:
                self.initialized = True
                if self._initialized_callback:
                    await self._initialized_callback()

        self._set_error(None)

    async def get_limit_switch_status(self) -> None:
        """Get the limit switch status."""
        status = await self._driver.get_limit_switches_status()
        self.limit_switch_status = {
            axis: StackerAxisState.from_status(status, axis) for axis in StackerAxis
        }

    async def get_motion_parameters(self) -> None:
        """Get the motion parameters used by the axis motors."""
        for axis in StackerAxis:
            self.motion_params[axis].move_params = await self._driver.get_motion_params(
                axis
            )

    async def get_platform_sensor_state(self) -> None:
        """Get the platform state."""
        status = await self._driver.get_platform_status()
        platform_state = PlatformState.from_status(status)
        if self.initialized and platform_state == PlatformState.UNKNOWN:
            # If the platform state is unknown but the X axis is known,
            # the platform is missing.
            await self.get_limit_switch_status()
            if self.limit_switch_status[StackerAxis.X] != StackerAxisState.UNKNOWN:
                platform_state = PlatformState.MISSING
        self.platform_state = platform_state

    async def get_door_closed(self) -> None:
        """Check if the hopper door is closed."""
        old_door_state = self.hopper_door_closed
        self.hopper_door_closed = await self._driver.get_hopper_door_closed()
        if old_door_state != self.hopper_door_closed and self._initialized_callback:
            await self._initialized_callback()

    async def get_installation_detected(self) -> None:
        """Check if the stacker install detect is set."""
        detected = await self._driver.get_installation_detected()
        self.installation_detected = detected

    def set_refresh_state(self) -> None:
        """Tell the reader to refresh all states, even ones that arent polled."""
        self._refresh_state = True

    def on_error(self, exception: Exception) -> None:
        self._driver.reset_serial_buffers()
        self._set_error(exception)

    def _set_error(self, exception: Optional[Exception]) -> None:
        if exception is None:
            self.error = None
        else:
            if self._error_callback:
                self._error_callback(exception)
            try:
                self.error = str(exception.args[0])
            except Exception:
                self.error = repr(exception)



import typing
from typing import Callable, Generator, List, Optional
from unittest import mock

import pytest

from opentrons_shared_data.protocol.constants import (
    JsonMagneticModuleCommand as JMMC,
)
from opentrons_shared_data.protocol.constants import (
    JsonPipetteCommand as JPC,
)
from opentrons_shared_data.protocol.constants import (
    JsonTemperatureModuleCommand as JTMC,
)
from opentrons_shared_data.protocol.constants import (
    JsonThermocyclerCommand as JTHC,
)
from opentrons_shared_data.protocol.types import (
    JsonProtocolV4,
    MagneticModuleEngageParams,
    ModuleIDParams,
    TemperatureParams,
    ThermocyclerAwaitBlockTemperatureCommand,
    ThermocyclerAwaitLidTemperatureCommand,
    ThermocyclerAwaitProfileCommand,
    ThermocyclerCommand,
    ThermocyclerRunProfileCommand,
    ThermocyclerRunProfileParams,
    ThermocyclerSetTargetBlockCommand,
    ThermocyclerSetTargetBlockParams,
    ThermocyclerSetTargetLidCommand,
)

import opentrons.protocols.execution.execute_json_v4 as v4
from opentrons.protocol_api import (
    MagneticModuleContext,
    ProtocolContext,
    TemperatureModuleContext,
    ThermocyclerContext,
)
from opentrons.protocols.execution import execute
from opentrons.protocols.execution.dev_types import (
    JsonV4MagneticModuleDispatch,
    JsonV4TemperatureModuleDispatch,
    JsonV4ThermocyclerDispatch,
    PipetteDispatch,
)
from opentrons.protocols.execution.execute_json_v4 import (
    TC_SPANNING_SLOT,
    _disengage_magnet,
    _engage_magnet,
    _temperature_module_await_temp,
    _temperature_module_deactivate,
    _temperature_module_set_temp,
    _thermocycler_close_lid,
    _thermocycler_deactivate_block,
    _thermocycler_deactivate_lid,
    _thermocycler_open_lid,
    _thermocycler_run_profile,
    _thermocycler_set_block_temperature,
    _thermocycler_set_lid_temperature,
    assert_no_async_tc_behavior,
    assert_tc_commands_do_not_use_unimplemented_params,
    dispatch_json,
    load_modules_from_json,
)
from opentrons.protocols.parse import parse


# autouse set to True to setup/teardown mock after each run
@pytest.fixture(autouse=True)
def mockObj() -> Generator[mock.Mock, None, None]:
    m = mock.Mock()
    yield m
    m.reset()


@pytest.fixture
def pipette_command_map(mockObj: mock.Mock) -> "PipetteDispatch":
    mock_pipette_command_map: "PipetteDispatch" = {
        JPC.blowout.value: mockObj._blowout,
        JPC.pickUpTip.value: mockObj._pick_up_tip,
        JPC.dropTip.value: mockObj._drop_tip,
        JPC.aspirate.value: mockObj._aspirate,
        JPC.dispense.value: mockObj._dispense,
        JPC.touchTip.value: mockObj._touch_tip,
    }
    return mock_pipette_command_map


@pytest.fixture
def magnetic_module_command_map(mockObj: mock.Mock) -> "JsonV4MagneticModuleDispatch":
    mock_magnetic_module_command_map: "JsonV4MagneticModuleDispatch" = {
        JMMC.magneticModuleEngageMagnet.value: mockObj._engage_magnet,
        JMMC.magneticModuleDisengageMagnet.value: mockObj._disengage_magnet,
    }

    return mock_magnetic_module_command_map


@pytest.fixture
def temperature_module_command_map(
    mockObj: mock.Mock,
) -> "JsonV4TemperatureModuleDispatch":
    mock_temperature_module_command_map: "JsonV4TemperatureModuleDispatch" = {
        JTMC.temperatureModuleSetTargetTemperature.value: mockObj._temperature_module_set_temp,
        JTMC.temperatureModuleDeactivate.value: mockObj._temperature_module_deactivate,
        JTMC.temperatureModuleAwaitTemperature.value: mockObj._temperature_module_await_temp,
    }
    return mock_temperature_module_command_map


@pytest.fixture
def thermocycler_module_command_map(mockObj: mock.Mock) -> "JsonV4ThermocyclerDispatch":
    mock_thermocycler_module_command_map: "JsonV4ThermocyclerDispatch" = {
        JTHC.thermocyclerCloseLid.value: mockObj._thermocycler_close_lid,
        JTHC.thermocyclerOpenLid.value: mockObj._thermocycler_open_lid,
        JTHC.thermocyclerDeactivateBlock.value: mockObj._thermocycler_deactivate_block,
        JTHC.thermocyclerDeactivateLid.value: mockObj._thermocycler_deactivate_lid,
        JTHC.thermocyclerSetTargetBlockTemperature.value: mockObj._thermocycler_set_block_temperature,
        JTHC.thermocyclerSetTargetLidTemperature.value: mockObj._thermocycler_set_lid_temperature,
        JTHC.thermocyclerRunProfile.value: mockObj._thermocycler_run_profile,
        # NOTE: the thermocyclerAwaitX commands are expected to always
        # follow a corresponding SetX command, which is implemented as
        # blocking. Then nothing needs to be done for awaitX commands.
        JTHC.thermocyclerAwaitBlockTemperature.value: mockObj.tc_do_nothing,
        JTHC.thermocyclerAwaitLidTemperature.value: mockObj.tc_do_nothing,
        JTHC.thermocyclerAwaitProfileComplete.value: mockObj.tc_do_nothing,
    }
    return mock_thermocycler_module_command_map


def test_load_modules_from_json() -> None:
    def fake_module(model: str, slot: Optional[str] = None) -> str:
        return "mock_module_ctx_" + model

    ctx = mock.create_autospec(ProtocolContext)
    ctx.load_module.side_effect = fake_module
    protocol: "JsonProtocolV4" = {
        "modules": {
            "aID": {"slot": "1", "model": "magneticModuleV1"},
            "bID": {"slot": "4", "model": "temperatureModuleV2"},
            "cID": {"slot": TC_SPANNING_SLOT, "model": "thermocyclerModuleV1"},
        }
    }
    result = load_modules_from_json(ctx, protocol)

    # load_module should be called in a deterministic order
    assert ctx.mock_calls == [
        mock.call.load_module("magneticModuleV1", "1"),
        mock.call.load_module("temperatureModuleV2", "4"),
        mock.call.load_module("thermocyclerModuleV1"),
    ]

    # resulting dict should have ModuleContext objects (from calling
    # load_module) as its values
    assert result == {  # type: ignore[comparison-overlap]
        "aID": "mock_module_ctx_magneticModuleV1",
        "bID": "mock_module_ctx_temperatureModuleV2",
        "cID": "mock_module_ctx_thermocyclerModuleV1",
    }


def test_engage_magnet() -> None:
    module_mock = mock.create_autospec(MagneticModuleContext)
    params: "MagneticModuleEngageParams" = {
        "module": "someModuleId",
        "engageHeight": 4.2,
    }
    _engage_magnet(module_mock, params)

    assert module_mock.mock_calls == [mock.call.engage(height_from_base=4.2)]


def test_disengage_magnet() -> None:
    module_mock = mock.create_autospec(MagneticModuleContext)
    params: "ModuleIDParams" = {"module": "someModuleId"}
    _disengage_magnet(module_mock, params)

    assert module_mock.mock_calls == [mock.call.disengage()]


def test_temperature_module_set_temp() -> None:
    module_mock = mock.create_autospec(TemperatureModuleContext)
    params: "TemperatureParams" = {"module": "someModuleId", "temperature": 42.5}
    _temperature_module_set_temp(module_mock, params)

    assert module_mock.mock_calls == [mock.call.start_set_temperature(42.5)]


def test_temperature_module_deactivate() -> None:
    module_mock = mock.create_autospec(TemperatureModuleContext)
    params: "ModuleIDParams" = {"module": "someModuleId"}
    _temperature_module_deactivate(module_mock, params)

    assert module_mock.mock_calls == [mock.call.deactivate()]


def test_temperature_module_await_temp() -> None:
    module_mock = mock.create_autospec(TemperatureModuleContext)
    params: "TemperatureParams" = {"module": "someModuleId", "temperature": 12.3}
    _temperature_module_await_temp(module_mock, params)

    assert module_mock.mock_calls == [mock.call.await_temperature(12.3)]


def test_thermocycler_close_lid() -> None:
    module_mock = mock.create_autospec(ThermocyclerContext)
    params: "ModuleIDParams" = {"module": "someModuleId"}
    _thermocycler_close_lid(module_mock, params)
    assert module_mock.mock_calls == [mock.call.close_lid()]


def test_thermocycler_open_lid() -> None:
    module_mock = mock.create_autospec(ThermocyclerContext)
    params: "ModuleIDParams" = {"module": "someModuleId"}
    _thermocycler_open_lid(module_mock, params)
    assert module_mock.mock_calls == [mock.call.open_lid()]


def test_thermocycler_deactivate_block() -> None:
    module_mock = mock.create_autospec(ThermocyclerContext)
    params: "ModuleIDParams" = {"module": "someModuleId"}
    _thermocycler_deactivate_block(module_mock, params)
    assert module_mock.mock_calls == [mock.call.deactivate_block()]


def test_thermocycler_deactivate_lid() -> None:
    module_mock = mock.create_autospec(ThermocyclerContext)
    params: "ModuleIDParams" = {"module": "someModuleId"}
    _thermocycler_deactivate_lid(module_mock, params)
    assert module_mock.mock_calls == [mock.call.deactivate_lid()]


def test_thermocycler_set_block_temperature() -> None:
    module_mock = mock.create_autospec(ThermocyclerContext)
    params: "ThermocyclerSetTargetBlockParams" = {
        "temperature": 42,
        "module": "someModuleId",
    }
    _thermocycler_set_block_temperature(module_mock, params)
    assert module_mock.mock_calls == [mock.call.set_block_temperature(42)]


def test_thermocycler_set_lid_temperature() -> None:
    module_mock = mock.create_autospec(ThermocyclerContext)
    params: "TemperatureParams" = {"module": "someModuleId", "temperature": 42}
    _thermocycler_set_lid_temperature(module_mock, params)
    assert module_mock.mock_calls == [mock.call.set_lid_temperature(42)]


def test_thermocycler_run_profile() -> None:
    module_mock = mock.create_autospec(ThermocyclerContext)
    params: "ThermocyclerRunProfileParams" = {
        "profile": [
            {"temperature": 55, "holdTime": 90},
            {"temperature": 65, "holdTime": 30},
        ],
        "module": "someModuleId",
        "volume": 98,
    }

    steps = [
        {"temperature": 55, "hold_time_seconds": 90},
        {"temperature": 65, "hold_time_seconds": 30},
    ]
    _thermocycler_run_profile(module_mock, params)
    assert module_mock.mock_calls == [
        mock.call.execute_profile(steps=steps, block_max_volume=98, repetitions=1)
    ]


def test_dispatch_json(
    monkeypatch: pytest.MonkeyPatch,
    pipette_command_map: "PipetteDispatch",
    magnetic_module_command_map: "JsonV4MagneticModuleDispatch",
    temperature_module_command_map: "JsonV4TemperatureModuleDispatch",
    thermocycler_module_command_map: "JsonV4ThermocyclerDispatch",
    mockObj: mock.Mock,
) -> None:
    monkeypatch.setattr(v4, "_delay", mockObj)
    monkeypatch.setattr(v4, "_move_to_slot", mockObj)

    magnetic_module_id = "magnetic_module_id"
    temperature_module_id = "temperature_module_id"
    thermocycler_module_id = "thermocycler_module_id"

    mock_magnetic_module = mock.create_autospec(MagneticModuleContext)
    mock_temperature_module = mock.create_autospec(TemperatureModuleContext)
    mock_thermocycler_module = mock.create_autospec(ThermocyclerContext)

    protocol_data = {
        "commands": [
            # Pipette
            {"command": "delay", "params": "delay_params"},
            {"command": "blowout", "params": "blowout_params"},
            {"command": "pickUpTip", "params": "pickUpTip_params"},
            {"command": "dropTip", "params": "dropTip_params"},
            {"command": "aspirate", "params": "aspirate_params"},
            {"command": "dispense", "params": "dispense_params"},
            {"command": "touchTip", "params": "touchTip_params"},
            {"command": "moveToSlot", "params": "moveToSlot_params"},
            # Magnetic Module
            {
                "command": "magneticModule/engageMagnet",
                "params": {"module": magnetic_module_id},
            },
            {
                "command": "magneticModule/disengageMagnet",
                "params": {"module": magnetic_module_id},
            },
            # Temperature Module
            {
                "command": "temperatureModule/setTargetTemperature",
                "params": {"module": temperature_module_id},
            },
            {
                "command": "temperatureModule/deactivate",
                "params": {"module": temperature_module_id},
            },
            {
                "command": "temperatureModule/awaitTemperature",
                "params": {"module": temperature_module_id},
            },
            # Thermocycler
            {
                "command": "thermocycler/setTargetBlockTemperature",
                "params": {"module": thermocycler_module_id, "temperature": 55},
            },
            {
                "command": "thermocycler/awaitBlockTemperature",
                "params": {"module": thermocycler_module_id, "temperature": 55},
            },
            {
                "command": "thermocycler/setTargetLidTemperature",
                "params": {"module": thermocycler_module_id, "temperature": 60},
            },
            {
                "command": "thermocycler/awaitLidTemperature",
                "params": {"module": thermocycler_module_id, "temperature": 60},
            },
            {
                "command": "thermocycler/closeLid",
                "params": {"module": thermocycler_module_id},
            },
            {
                "command": "thermocycler/openLid",
                "params": {"module": thermocycler_module_id},
            },
            {
                "command": "thermocycler/deactivateBlock",
                "params": {"module": thermocycler_module_id},
            },
            {
                "command": "thermocycler/deactivateLid",
                "params": {"module": thermocycler_module_id},
            },
            {
                "command": "thermocycler/closeLid",
                "params": {"module": thermocycler_module_id},
            },
            {
                "command": "thermocycler/runProfile",
                "params": {
                    "module": thermocycler_module_id,
                    "profile": [
                        {"temperature": 70, "holdTime": 60},
                        {"temperature": 40, "holdTime": 30},
                        {"temperature": 72, "holdTime": 60},
                        {"temperature": 38, "holdTime": 30},
                    ],
                    "volume": 123,
                },
            },
            {
                "command": "thermocycler/awaitProfileComplete",
                "params": {"module": thermocycler_module_id},
            },
        ]
    }

    context = mock.sentinel.context
    instruments = mock.sentinel.instruments
    loaded_labware = mock.sentinel.loaded_labware

    modules = {
        magnetic_module_id: mock_magnetic_module,
        temperature_module_id: mock_temperature_module,
        thermocycler_module_id: mock_thermocycler_module,
    }

    dispatch_json(
        context,
        protocol_data,  # type: ignore[arg-type]
        instruments,
        loaded_labware,
        modules,
        pipette_command_map,
        magnetic_module_command_map,
        temperature_module_command_map,
        thermocycler_module_command_map,
    )

    assert mockObj.mock_calls == [
        # Pipette
        mock.call._delay(context, "delay_params"),
        mock.call._blowout(instruments, loaded_labware, "blowout_params"),
        mock.call._pick_up_tip(instruments, loaded_labware, "pickUpTip_params"),
        mock.call._drop_tip(instruments, loaded_labware, "dropTip_params"),
        mock.call._aspirate(instruments, loaded_labware, "aspirate_params"),
        mock.call._dispense(instruments, loaded_labware, "dispense_params"),
        mock.call._touch_tip(instruments, loaded_labware, "touchTip_params"),
        mock.call._move_to_slot(context, instruments, "moveToSlot_params"),
        # Magnetic module
        mock.call._engage_magnet(mock_magnetic_module, {"module": magnetic_module_id}),
        mock.call._disengage_magnet(
            mock_magnetic_module, {"module": magnetic_module_id}
        ),
        # Temperature module
        mock.call._temperature_module_set_temp(
            mock_temperature_module, {"module": temperature_module_id}
        ),
        mock.call._temperature_module_deactivate(
            mock_temperature_module, {"module": temperature_module_id}
        ),
        mock.call._temperature_module_await_temp(
            mock_temperature_module, {"module": temperature_module_id}
        ),
        # Thermocycler
        mock.call._thermocycler_set_block_temperature(
            mock_thermocycler_module,
            {"module": thermocycler_module_id, "temperature": 55},
        ),
        # await block temp = do nothing (corresponding set is blocking)
        mock.call.tc_do_nothing(
            mock_thermocycler_module,
            {"module": thermocycler_module_id, "temperature": 55},
        ),
        mock.call._thermocycler_set_lid_temperature(
            mock_thermocycler_module,
            {"module": thermocycler_module_id, "temperature": 60},
        ),
        # await lid temp = do nothing (corresponding set is blocking)
        mock.call.tc_do_nothing(
            mock_thermocycler_module,
            {"module": thermocycler_module_id, "temperature": 60},
        ),
        mock.call._thermocycler_close_lid(
            mock_thermocycler_module, {"module": thermocycler_module_id}
        ),
        mock.call._thermocycler_open_lid(
            mock_thermocycler_module, {"module": thermocycler_module_id}
        ),
        mock.call._thermocycler_deactivate_block(
            mock_thermocycler_module, {"module": thermocycler_module_id}
        ),
        mock.call._thermocycler_deactivate_lid(
            mock_thermocycler_module, {"module": thermocycler_module_id}
        ),
        mock.call._thermocycler_close_lid(
            mock_thermocycler_module, {"module": thermocycler_module_id}
        ),
        mock.call._thermocycler_run_profile(
            mock_thermocycler_module,
            {
                "module": thermocycler_module_id,
                "profile": [
                    {"temperature": 70, "holdTime": 60},
                    {"temperature": 40, "holdTime": 30},
                    {"temperature": 72, "holdTime": 60},
                    {"temperature": 38, "holdTime": 30},
                ],
                "volume": 123,
            },
        ),
        # await profile complete = do nothing
        # (corresponding set is blocking)
        mock.call.tc_do_nothing(
            mock_thermocycler_module, {"module": thermocycler_module_id}
        ),
    ]


@typing.no_type_check
def test_dispatch_json_invalid_command(
    pipette_command_map: "PipetteDispatch",
    magnetic_module_command_map: "JsonV4MagneticModuleDispatch",
    temperature_module_command_map: "JsonV4TemperatureModuleDispatch",
    thermocycler_module_command_map: "JsonV4ThermocyclerDispatch",
) -> None:
    protocol_data: "JsonProtocolV4" = {
        "commands": [
            {"command": "no_such_command", "params": "foo"},
        ]
    }
    with pytest.raises(RuntimeError):
        dispatch_json(
            context=None,
            protocol_data=protocol_data,
            instruments=None,
            loaded_labware=None,
            modules=None,
            pipette_command_map=pipette_command_map,
            magnetic_module_command_map=magnetic_module_command_map,
            temperature_module_command_map=temperature_module_command_map,
            thermocycler_module_command_map=thermocycler_module_command_map,
        )


@pytest.mark.ot2_only
def test_papi_execute_json_v4(
    monkeypatch: pytest.MonkeyPatch,
    ctx: ProtocolContext,
    get_json_protocol_fixture: Callable[[str, str, bool], str],
) -> None:
    protocol_data = get_json_protocol_fixture("4", "testModulesProtocol", False)
    protocol = parse(protocol_data, None)
    ctx.home()
    # Check that we end up executing the protocol ok
    execute.run_protocol(protocol, ctx)


def test_assert_no_async_tc_behavior() -> None:
    temp_params: "TemperatureParams" = {
        "temperature": 75,
        "module": "thermocycler_module_id",
    }
    set_target_params: "ThermocyclerSetTargetBlockParams" = {
        "volume": 100,
        **temp_params,
    }
    run_profile_params: "ThermocyclerRunProfileParams" = {
        "module": "thermocycler_module_id",
        "volume": 100,
        "profile": [],
    }

    setBlock: "ThermocyclerSetTargetBlockCommand" = {
        "command": "thermocycler/setTargetBlockTemperature",
        "params": set_target_params,
    }
    awaitBlock: "ThermocyclerAwaitBlockTemperatureCommand" = {
        "command": "thermocycler/awaitBlockTemperature",
        "params": temp_params,
    }
    setLid: "ThermocyclerSetTargetLidCommand" = {
        "command": "thermocycler/setTargetLidTemperature",
        "params": temp_params,
    }
    awaitLid: "ThermocyclerAwaitLidTemperatureCommand" = {
        "command": "thermocycler/awaitLidTemperature",
        "params": temp_params,
    }
    runProfile: "ThermocyclerRunProfileCommand" = {
        "command": "thermocycler/runProfile",
        "params": run_profile_params,
    }
    awaitProfile: "ThermocyclerAwaitProfileCommand" = {
        "command": "thermocycler/awaitProfileComplete",
        "params": run_profile_params,
    }
    anything: "ThermocyclerCommand" = {"command": "foo", "params": {}}  # type: ignore[assignment, misc]  # stand-in for some other command

    # no error
    assert_no_async_tc_behavior(
        [
            anything,
            setBlock,
            awaitBlock,
            anything,
            setLid,
            awaitLid,
            anything,
            runProfile,
            awaitProfile,
            anything,
        ]
    )

    assert_no_async_tc_behavior([anything, anything])

    setters: List[ThermocyclerCommand] = [setBlock, setLid, runProfile]

    for setter in setters:
        # Should raise if anything except the corresponding await
        # follows a "setter" command
        with pytest.raises(RuntimeError):
            assert_no_async_tc_behavior([setter, anything])

        # Should raise if setter is the last command in the array
        with pytest.raises(RuntimeError):
            assert_no_async_tc_behavior([anything, setter])

    awaiters: List[ThermocyclerCommand] = [awaitBlock, awaitLid, awaitProfile]

    for awaiter in awaiters:
        # Should raise if anything except the corresponding set
        # precedes the await
        with pytest.raises(RuntimeError):
            assert_no_async_tc_behavior([anything, awaiter])

        # Should raise if awaiter is the first command in the array
        with pytest.raises(RuntimeError):
            assert_no_async_tc_behavior([awaiter, anything])


def test_assert_tc_commands_do_not_use_unimplemented_params() -> None:
    # should not throw for arbitrary commands
    assert_tc_commands_do_not_use_unimplemented_params(
        [{"command": "foo", "params": {}}]  # type: ignore[list-item, misc]
    )

    fail_cases: List[List[ThermocyclerSetTargetBlockCommand]] = [
        [
            {
                "command": "thermocycler/setTargetBlockTemperature",
                "params": {
                    "volume": 0,
                    "temperature": 0,
                    "module": "thermocycler_module_id",
                },
            }
        ]
    ]
    for cmds in fail_cases:
        with pytest.raises(RuntimeError):
            assert_tc_commands_do_not_use_unimplemented_params(cmds)  # type: ignore[arg-type]



from __future__ import annotations

from typing import Callable, Literal, NamedTuple, Optional, Tuple, TypedDict


class APIVersion(NamedTuple):
    major: int
    minor: int

    @classmethod
    def from_string(cls, inp: str) -> APIVersion:
        parts = inp.split(".")
        if len(parts) != 2:
            raise ValueError(inp)
        intparts = [int(p) for p in parts]

        return cls(major=intparts[0], minor=intparts[1])

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}"


class TransferArgs(TypedDict, total=False):
    """
    The common arguments for the transfer functions `InstrumentContext.transfer()`,
    `InstrumentContext.distribute()`, `InstrumentContext.consolidate()`.
    """

    mode: Literal["transfer", "distribute", "consolidate"]  # internal use only
    new_tip: Literal["once", "always", "never"]
    trash: bool
    touch_tip: bool
    blow_out: bool
    blowout_location: Literal["trash", "source well", "destination well"]
    mix_before: Tuple[int, float]
    mix_after: Tuple[int, float]
    disposal_volume: float
    air_gap: float
    carryover: bool  # this does nothing!
    gradient_function: Callable[[float], float]  # very mysterious


class ThermocyclerStepBase(TypedDict):
    """Required elements of a thermocycler step: the temperature."""

    temperature: float


class ThermocyclerStep(ThermocyclerStepBase, total=False):
    """Optional elements of a thermocycler step: the hold time. One of these must be present."""

    hold_time_seconds: float
    hold_time_minutes: float
    ramp_rate: Optional[float]



"""Tests for `LegacyBroker`."""

from typing import List, NamedTuple, cast

from opentrons.legacy_commands.publisher import CommandPublisher, publish
from opentrons.legacy_commands.types import CommandMessage


def _my_command(arg1: int, arg2: str = "", arg3: str = "") -> CommandMessage:
    return cast(
        CommandMessage,
        {
            "name": "command",
            "payload": {"description": f"arg1:{arg1}, arg2:{arg2}, arg3:{arg3}"},
        },
    )


class _Call(NamedTuple):
    level: int
    description: str


class _FakeClass(CommandPublisher):
    def __init__(self) -> None:
        super().__init__(None)

    @publish(command=_my_command)
    def method_a(self, arg1: int, arg2: str, arg3: str = "foo") -> int:
        self.method_b(0)
        return 100

    @publish(command=_my_command)
    def method_b(self, arg1: int) -> None:
        return None

    @publish(command=_my_command)
    def method_c(self, arg1: int, arg2: str, arg3: str = "bar") -> int:
        self.method_b(0)
        return 100


def test_legacy_broker() -> None:
    """Test subscribing, emitting, receiving, and unsubscribing."""
    stack: List[CommandMessage] = []
    calls: List[_Call] = []
    fake_obj = _FakeClass()

    def on_notify(message: CommandMessage) -> None:
        assert message["name"] == "command"  # type: ignore[comparison-overlap]
        payload = message["payload"]
        description = payload["description"]

        if message["$"] == "before":
            stack.append(message)
            calls.append(_Call(level=len(stack), description=description))
        else:
            stack.pop()

    unsubscribe = fake_obj.broker.subscribe("command", on_notify)

    fake_obj.method_a(0, "1")
    fake_obj.method_b(2)
    fake_obj.method_c(3, "4")

    expected = [
        _Call(level=1, description="arg1:0, arg2:1, arg3:foo"),
        _Call(level=2, description="arg1:0, arg2:, arg3:"),
        _Call(level=1, description="arg1:2, arg2:, arg3:"),
        _Call(level=1, description="arg1:3, arg2:4, arg3:bar"),
        _Call(level=2, description="arg1:0, arg2:, arg3:"),
    ]

    assert calls == expected

    unsubscribe()
    fake_obj.method_a(0, "2")

    assert calls == expected, "No calls expected after unsubscribe()"



"""Flex Stacker substate."""

from dataclasses import dataclass
from typing import NewType

from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.protocol_engine.state.update_types import (
    NO_CHANGE,
    FlexStackerStateUpdate,
)
from opentrons.protocol_engine.types.module import (
    StackerPoolDefinition,
    StackerStoredLabwareGroup,
)

FlexStackerId = NewType("FlexStackerId", str)


@dataclass(frozen=True)
class FlexStackerSubState:
    """Flex Stacker-specific state.

    Provides calculations and read-only state access
    for an individual loaded Flex Stacker Module.
    """

    module_id: FlexStackerId
    pool_primary_definition: LabwareDefinition | None
    pool_adapter_definition: LabwareDefinition | None
    pool_lid_definition: LabwareDefinition | None
    max_pool_count: int
    pool_height: float
    pool_overlap: float
    contained_labware_bottom_first: list[StackerStoredLabwareGroup]

    def new_from_state_change(
        self, update: FlexStackerStateUpdate
    ) -> "FlexStackerSubState":
        """Return a new state with the given update applied."""
        pool_primary_definition = self.pool_primary_definition
        pool_adapter_definition = self.pool_adapter_definition
        pool_lid_definition = self.pool_lid_definition
        max_pool_count = self.max_pool_count
        pool_overlap = self.pool_overlap
        pool_height = self.pool_height
        if update.pool_constraint != NO_CHANGE:
            max_pool_count = update.pool_constraint.max_pool_count
            pool_primary_definition = update.pool_constraint.primary_definition
            pool_adapter_definition = update.pool_constraint.adapter_definition
            pool_lid_definition = update.pool_constraint.lid_definition
            pool_overlap = update.pool_constraint.pool_overlap
            pool_height = update.pool_constraint.pool_height

        contained_labware = self.contained_labware_bottom_first

        if update.contained_labware_bottom_first != NO_CHANGE:
            contained_labware = update.contained_labware_bottom_first

        return FlexStackerSubState(
            module_id=self.module_id,
            pool_primary_definition=pool_primary_definition,
            pool_adapter_definition=pool_adapter_definition,
            pool_lid_definition=pool_lid_definition,
            contained_labware_bottom_first=contained_labware,
            max_pool_count=max_pool_count,
            pool_overlap=pool_overlap,
            pool_height=pool_height,
        )

    def get_pool_definition_ordered_list(self) -> list[LabwareDefinition] | None:
        """Get the pool definitions in a list suitable for getting the height."""
        if not self.pool_primary_definition:
            return None

        defs: list[LabwareDefinition] = []
        if self.pool_lid_definition is not None:
            defs.append(self.pool_lid_definition)
        defs.append(self.pool_primary_definition)
        if self.pool_adapter_definition is not None:
            defs.append(self.pool_adapter_definition)
        return defs

    def get_pool_height_minus_overlap(self) -> float:
        """Get the height used in dispense/store action."""
        return self.pool_height - self.pool_overlap

    def get_contained_labware(self) -> list[StackerStoredLabwareGroup]:
        """Get the labware inside the hopper."""
        return self.contained_labware_bottom_first

    def get_max_pool_count(self) -> int | None:
        """Get the maximum number of currently-configured labware.

        If the stacker has not been configured, return None.
        """
        if not self.pool_primary_definition:
            return None
        return self.max_pool_count

    def get_pool_overlap(self) -> float:
        """Get the overlap of the currently-configured labware."""
        return self.pool_overlap

    def get_pool_definition(self) -> StackerPoolDefinition | None:
        """Get the labware definitions of the stacker pool."""
        if not self.pool_primary_definition:
            return None
        return StackerPoolDefinition(
            primaryLabwareDefinition=self.pool_primary_definition,
            adapterLabwareDefinition=self.pool_adapter_definition,
            lidLabwareDefinition=self.pool_lid_definition,
        )



from typing import Dict, Generator, Optional
from unittest.mock import MagicMock, patch

import pytest
from pytest_lazy_fixtures import lf as lazy_fixture

from opentrons_shared_data.robot.types import RobotTypeEnum

from opentrons.config import CONFIG, advanced_settings


@pytest.fixture
def mock_settings_values_ot2() -> Dict[str, Optional[bool]]:
    return {
        s.id: False
        for s in advanced_settings.settings
        if RobotTypeEnum.OT2 in s.robot_type and not s.internal_only
    }


@pytest.fixture
def mock_settings_values_ot2_all() -> Dict[str, Optional[bool]]:
    return {
        s.id: False
        for s in advanced_settings.settings
        if RobotTypeEnum.OT2 in s.robot_type
    }


@pytest.fixture
def mock_settings_values_flex() -> Dict[str, Optional[bool]]:
    return {
        s.id: False
        for s in advanced_settings.settings
        if RobotTypeEnum.FLEX in s.robot_type and not s.internal_only
    }


@pytest.fixture
def mock_settings_values_flex_all() -> Dict[str, Optional[bool]]:
    return {
        s.id: False
        for s in advanced_settings.settings
        if RobotTypeEnum.FLEX in s.robot_type
    }


@pytest.fixture
def mock_settings_values_empty() -> Dict[str, Optional[bool]]:
    return {s.id: None for s in advanced_settings.settings}


@pytest.fixture
def mock_settings_version() -> int:
    return 1


@pytest.fixture
def mock_settings(
    mock_settings_values_ot2: Dict[str, Optional[bool]],
    mock_settings_version: int,
) -> advanced_settings.SettingsData:
    return advanced_settings.SettingsData(
        settings_map=mock_settings_values_ot2,
        version=mock_settings_version,
    )


@pytest.fixture
def mock_read_settings_file_ot2(
    mock_settings_values_ot2_all: Dict[str, Optional[bool]],
    mock_settings_version: int,
) -> Generator[MagicMock, None, None]:
    with patch("opentrons.config.advanced_settings._read_settings_file") as p:
        p.return_value = advanced_settings.SettingsData(
            settings_map=mock_settings_values_ot2_all,
            version=mock_settings_version,
        )
        yield p


@pytest.fixture
def mock_read_settings_file_flex(
    mock_settings_values_flex_all: Dict[str, Optional[bool]],
    mock_settings_version: int,
) -> Generator[MagicMock, None, None]:
    with patch("opentrons.config.advanced_settings._read_settings_file") as p:
        p.return_value = advanced_settings.SettingsData(
            settings_map=mock_settings_values_flex_all,
            version=mock_settings_version,
        )
        yield p


@pytest.fixture
def mock_read_settings_file_empty(
    mock_settings_values_empty: Dict[str, Optional[bool]],
    mock_settings_version: int,
) -> Generator[MagicMock, None, None]:
    with patch("opentrons.config.advanced_settings._read_settings_file") as p:
        p.return_value = advanced_settings.SettingsData(
            settings_map=mock_settings_values_empty,
            version=mock_settings_version,
        )
        yield p


@pytest.fixture
def mock_write_settings_file() -> Generator[MagicMock, None, None]:
    with patch("opentrons.config.advanced_settings._write_settings_file") as p:
        yield p


@pytest.fixture
def restore_restart_required() -> Generator[None, None, None]:
    yield
    advanced_settings._SETTINGS_RESTART_REQUIRED = False


@pytest.fixture
def clear_cache() -> None:
    advanced_settings.get_all_adv_settings.cache_clear()


def test_get_advanced_setting_not_found(
    clear_cache: None, mock_read_settings_file_ot2: MagicMock
) -> None:
    assert advanced_settings.get_adv_setting("unknown", RobotTypeEnum.OT2) is None


def test_get_advanced_setting_found(
    clear_cache: None,
    mock_read_settings_file_ot2: MagicMock,
    mock_settings_values_ot2_all: Dict[str, Optional[bool]],
) -> None:
    for k, v in mock_settings_values_ot2_all.items():
        s = advanced_settings.get_adv_setting(k, RobotTypeEnum.OT2)
        assert s is not None
        assert s.value == v
        assert s.definition == advanced_settings.settings_by_id[k]


@pytest.mark.parametrize(
    "robot_type, mock_settings_values, mock_read_settings_file",
    [
        [
            RobotTypeEnum.OT2,
            lazy_fixture("mock_settings_values_ot2"),
            lazy_fixture("mock_read_settings_file_ot2"),
        ],
        [
            RobotTypeEnum.FLEX,
            lazy_fixture("mock_settings_values_flex"),
            lazy_fixture("mock_read_settings_file_flex"),
        ],
    ],
)
def test_get_all_adv_settings(
    clear_cache: None,
    mock_read_settings_file: MagicMock,
    mock_settings_values: MagicMock,
    robot_type: RobotTypeEnum,
) -> None:
    s = advanced_settings.get_all_adv_settings(robot_type)
    assert s.keys() == mock_settings_values.keys()
    for k, v in s.items():
        assert v.value == mock_settings_values[k]
        assert v.definition == advanced_settings.settings_by_id[k]


def test_get_all_adv_settings_empty(
    clear_cache: None,
    mock_read_settings_file_ot2: MagicMock,
) -> None:
    mock_read_settings_file_ot2.return_value = advanced_settings.SettingsData({}, 1)
    s = advanced_settings.get_all_adv_settings(RobotTypeEnum.OT2)
    assert s == {}


async def test_set_adv_setting(
    mock_read_settings_file_ot2: MagicMock,
    mock_settings_values_ot2_all: MagicMock,
    mock_write_settings_file: MagicMock,
    mock_settings_version: int,
    restore_restart_required: None,
) -> None:
    for k, v in mock_settings_values_ot2_all.items():
        # Toggle the advanced setting
        await advanced_settings.set_adv_setting(k, not v)
        mock_write_settings_file.assert_called_with(
            # Only the current key is toggled
            {
                nk: nv if nk != k else not v
                for nk, nv in mock_settings_values_ot2_all.items()
            },
            mock_settings_version,
            CONFIG["feature_flags_file"],
        )


async def test_set_adv_setting_unknown(
    mock_read_settings_file_ot2: MagicMock,
    mock_write_settings_file: MagicMock,
) -> None:
    mock_read_settings_file_ot2.return_value = advanced_settings.SettingsData({}, 1)
    with pytest.raises(ValueError, match="is not recognized"):
        await advanced_settings.set_adv_setting("no", False)


async def test_get_all_adv_settings_lru_cache(
    clear_cache: None,
    mock_read_settings_file_ot2: MagicMock,
    mock_write_settings_file: MagicMock,
) -> None:
    # Cache should not be used.
    advanced_settings.get_all_adv_settings(RobotTypeEnum.OT2)
    mock_read_settings_file_ot2.assert_called_once()
    mock_read_settings_file_ot2.reset_mock()
    # Should use cache
    advanced_settings.get_all_adv_settings(RobotTypeEnum.OT2)
    mock_read_settings_file_ot2.assert_not_called()
    mock_read_settings_file_ot2.reset_mock()
    # Updating will invalidate cache
    await advanced_settings.set_adv_setting("enableDoorSafetySwitch", True)
    mock_read_settings_file_ot2.reset_mock()
    # Cache should not be used
    advanced_settings.get_all_adv_settings(RobotTypeEnum.OT2)
    mock_read_settings_file_ot2.assert_called_once()
    mock_read_settings_file_ot2.reset_mock()
    # Should use cache
    advanced_settings.get_all_adv_settings(RobotTypeEnum.OT2)
    mock_read_settings_file_ot2.assert_not_called()


async def test_restart_required(
    restore_restart_required: None,
    mock_read_settings_file_ot2: MagicMock,
    mock_write_settings_file: MagicMock,
    mock_settings_version: int,
) -> None:
    _id = "restart_required"
    # Mock out the available settings
    available_settings = [
        advanced_settings.SettingDefinition(
            _id=_id,
            title="",
            description="",
            restart_required=True,
            robot_type=[RobotTypeEnum.OT2, RobotTypeEnum.FLEX],
        )
    ]
    with patch.object(advanced_settings, "settings", new=available_settings):
        # Mock out the settings_by_id
        available_settings_by_id = {s.id: s for s in available_settings}
        with patch.object(
            advanced_settings, "settings_by_id", new=available_settings_by_id
        ):
            mock_read_settings_file_ot2.return_value = advanced_settings.SettingsData(
                settings_map={_id: None}, version=mock_settings_version
            )

            assert advanced_settings.is_restart_required() is False
            await advanced_settings.set_adv_setting(_id, True)
            assert advanced_settings.is_restart_required() is True


def test_per_robot_true_defaults(mock_read_settings_file_empty: MagicMock) -> None:
    with patch.object(advanced_settings, "settings_by_id", new={}):
        assert (
            advanced_settings.get_setting_with_env_overload(
                "enableDoorSafetySwitch", RobotTypeEnum.FLEX
            )
            is True
        )



"""Tip state tracking."""

from dataclasses import dataclass
from typing import Dict, Iterable, List, Optional, Set

from ..actions import Action, get_state_updates
from ._abstract_store import HandlesActions, HasState
from ._well_math import (
    wells_covered_by_physical_pipette,
    wells_covered_by_pipette_configuration,
    wells_covered_dense,
)
from opentrons.hardware_control.nozzle_manager import NozzleMap
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.types import TipRackWellState
from opentrons.types import NozzleConfigurationType, NozzleMapInterface

_TipRackStateByWellName = Dict[str, TipRackWellState]


@dataclass
class TipState:
    """State of all tips."""

    tips_by_labware_id: Dict[str, _TipRackStateByWellName]
    columns_by_labware_id: Dict[str, List[List[str]]]


class TipStore(HasState[TipState], HandlesActions):
    """Tip state container."""

    _state: TipState

    def __init__(self) -> None:
        """Initialize a liquid store and its state."""
        self._state = TipState(
            tips_by_labware_id={},
            columns_by_labware_id={},
        )

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        for state_update in get_state_updates(action):
            self._handle_state_update(state_update)

    def _handle_state_update(self, state_update: update_types.StateUpdate) -> None:
        if state_update.tips_state != update_types.NO_CHANGE:
            self._set_tip_state(
                labware_id=state_update.tips_state.labware_id,
                well_names=state_update.tips_state.well_names,
                tip_state=state_update.tips_state.tip_state,
            )

        if state_update.loaded_labware != update_types.NO_CHANGE:
            labware_id = state_update.loaded_labware.labware_id
            definition = state_update.loaded_labware.definition
            if definition.parameters.isTiprack:
                self._state.tips_by_labware_id[labware_id] = {
                    well_name: TipRackWellState.CLEAN
                    for column in definition.ordering
                    for well_name in column
                }
                self._state.columns_by_labware_id[labware_id] = [
                    column for column in definition.ordering
                ]
        if state_update.batch_loaded_labware != update_types.NO_CHANGE:
            for labware_id in state_update.batch_loaded_labware.new_locations_by_id:
                definition = state_update.batch_loaded_labware.definitions_by_id[
                    labware_id
                ]
                if definition.parameters.isTiprack:
                    self._state.tips_by_labware_id[labware_id] = {
                        well_name: TipRackWellState.CLEAN
                        for column in definition.ordering
                        for well_name in column
                    }
                    self._state.columns_by_labware_id[labware_id] = [
                        column for column in definition.ordering
                    ]

    def _set_tip_state(
        self, labware_id: str, well_names: Iterable[str], tip_state: TipRackWellState
    ) -> None:
        well_states = self._state.tips_by_labware_id.get(labware_id, {})
        for well_name in well_names:
            well_states[well_name] = tip_state


class TipView:
    """Read-only tip state view."""

    _state: TipState

    def __init__(self, state: TipState) -> None:
        """Initialize the computed view of liquid state.

        Arguments:
            state: Liquid state dataclass used for all calculations.
        """
        self._state = state

    def get_next_tip(  # noqa: C901
        self,
        labware_id: str,
        num_tips: int,
        starting_tip_name: Optional[str],
        nozzle_map: Optional[NozzleMapInterface],
    ) -> Optional[str]:
        """Gets the next available clean tip.

        Does not support use of a starting tip if the pipette used is in a partial configuration.
        """
        if starting_tip_name is None and nozzle_map is not None:
            return self._get_next_tip_with_nozzle_map(labware_id, nozzle_map)
        else:
            wells = self._state.tips_by_labware_id.get(labware_id, {})
            columns = self._state.columns_by_labware_id.get(labware_id, [])
            if columns and num_tips == len(columns[0]):  # Get next tips for 8-channel
                column_head = [column[0] for column in columns]
                starting_column_index = 0

                if starting_tip_name:
                    for idx, column in enumerate(columns):
                        if starting_tip_name in column:
                            if starting_tip_name not in column_head:
                                starting_column_index = idx + 1
                            else:
                                starting_column_index = idx

                for column in columns[starting_column_index:]:
                    if all(wells[well] == TipRackWellState.CLEAN for well in column):
                        return column[0]

            elif num_tips == len(wells.keys()):  # Get next tips for 96 channel
                if starting_tip_name and starting_tip_name != columns[0][0]:
                    return None

                if all(
                    tip_state == TipRackWellState.CLEAN for tip_state in wells.values()
                ):
                    return next(iter(wells))

            else:  # Get next tips for single channel
                if starting_tip_name is not None:
                    wells = _drop_wells_before_starting_tip(wells, starting_tip_name)

                for well_name, tip_state in wells.items():
                    if tip_state == TipRackWellState.CLEAN:
                        return well_name
        return None

    def _get_next_tip_with_nozzle_map(
        self,
        labware_id: str,
        nozzle_map: NozzleMapInterface,
    ) -> Optional[str]:
        """Get the next available clean tip for given nozzle configuration if one can be found."""
        tip_well_states = self._state.tips_by_labware_id.get(labware_id, {})
        wells_by_columns = self._state.columns_by_labware_id.get(labware_id, [])

        def _validate_wells(well_list: Set[str], target_well: str) -> bool:
            # If we are not picking up the correct number of tips it's not valid
            if len(well_list) != nozzle_map.tip_count:
                return False
            # If not all the tips we'll be picking up are clean it's not valid
            target_well_states = [tip_well_states[well_name] for well_name in well_list]
            if not all(state == TipRackWellState.CLEAN for state in target_well_states):
                return False
            # Since we know a full configuration will always produce zero non-active overlapping wells
            # we can skip the following checks if it is a full configuration.
            if nozzle_map.configuration != NozzleConfigurationType.FULL:
                # If we have a partial configuration we need to ensure that any wells in the way are NOT present
                wells_covered_physically = set(
                    wells_covered_by_physical_pipette(
                        nozzle_map=nozzle_map,  # type: ignore[arg-type]
                        target_well=target_well,
                        labware_wells_by_column=wells_by_columns,
                    )
                )
                wells_in_way_well_state = [
                    tip_well_states[well_name]
                    for well_name in wells_covered_physically.difference(well_list)
                ]
                # If any of the wells in the way are NOT empty, this is not a valid configuration
                if not all(
                    well_state == TipRackWellState.EMPTY
                    for well_state in wells_in_way_well_state
                ):
                    return False

            return True

        # Get an ordered list of wells to most efficiently search, depending on pipette configuration
        target_well_list = _resolve_well_order(wells_by_columns, nozzle_map)

        for well in target_well_list:
            # If the target well/tip isn't clean, skip to the next one. This will be checked
            # again in _validate_wells, but we can short circuit the following checks if this is False
            if tip_well_states[well] != TipRackWellState.CLEAN:
                continue
            # Get list of all wells (i.e. tips) that would be covered by the active nozzles
            targeted_wells = set(
                wells_covered_by_pipette_configuration(
                    nozzle_map=nozzle_map,  # type: ignore[arg-type]
                    target_well=well,
                    labware_wells_by_column=wells_by_columns,
                )
            )
            # If we are picking up the correct number of tips, return that target well
            if _validate_wells(targeted_wells, target_well=well):
                return well

        return None

    def has_clean_tip(self, labware_id: str, well_name: str) -> bool:
        """Get whether a well in a labware has a clean tip.

        Args:
            labware_id: The labware ID to check.
            well_name: The well name to check.

        Returns:
            True if the labware is a tip rack and the well has a clean tip,
            otherwise False.
        """
        tip_rack = self._state.tips_by_labware_id.get(labware_id)
        well_state = tip_rack.get(well_name) if tip_rack else None

        return well_state == TipRackWellState.CLEAN

    def compute_tips_to_mark_as_used_or_empty(
        self, labware_id: str, well_name: str, nozzle_map: NozzleMap
    ) -> list[str]:
        """Compute which tips a hypothetical tip pickup/drop should mark as "used" or "empty".

        Params:
            labware_id: The labware ID of the tip rack.
            well_name: The single target well of the tip pickup.
            nozzle_map: The nozzle configuration that the pipette will use for the pickup.

        Returns:
            The well names of all the tips that the operation will use.
        """
        columns = self._state.columns_by_labware_id.get(labware_id, [])
        return list(
            wells_covered_dense(
                nozzle_map.columns,
                nozzle_map.rows,
                nozzle_map.starting_nozzle,
                well_name,
                columns,
            )
        )


def _drop_wells_before_starting_tip(
    wells: _TipRackStateByWellName, starting_tip_name: str
) -> _TipRackStateByWellName:
    """Drop any wells that come before the starting tip and return the remaining ones after."""
    seen_starting_well = False
    remaining_wells: dict[str, TipRackWellState] = {}
    for well_name, tip_state in wells.items():
        if well_name == starting_tip_name:
            seen_starting_well = True
        if seen_starting_well:
            remaining_wells[well_name] = tip_state
    return remaining_wells


def _resolve_well_order(  # noqa: C901
    well_list: List[List[str]], nozzle_map: NozzleMapInterface
) -> List[str]:
    """Given a list of ordered columns and pipette information, returns a flat list of wells ordered for tip pick up.

    Wells can be ordered in four different ways:
        - Top to bottom, left to right (A1, B1, ... A2, B2, ... G12, H12)
        - Top to bottom, right to left (A12, B12, ... A11, B11, ... G1, H1)
        - Bottom to top, left to right (H1, G1, ... H2, G2, ... B12, A12)
        - Bottom to top, right to left (A12, B12, ... A11, B11, ... G1, H1)

    - Full configurations (which will always cover a single channel) will go top to bottom, left to right.
    - A partial 8-channel pipette configuration will always search left to right, starting at either top to bottom for
        starting nozzle H1 or bottom to top for starting nozzle A1
    - A partial 96-channel pipette configuration will always begin in the opposite corner of the starting nozzle
    """
    if nozzle_map.configuration == NozzleConfigurationType.FULL:
        return _get_top_to_bottom_left_to_right(well_list)
    elif nozzle_map.physical_nozzle_count == 8:
        if nozzle_map.starting_nozzle == "A1":
            return _get_bottom_to_top_left_to_right(well_list)
        elif nozzle_map.starting_nozzle == "H1":
            return _get_top_to_bottom_left_to_right(well_list)
        else:
            raise ValueError(
                f"Nozzle {nozzle_map.starting_nozzle} is an invalid starting tip for"
                " 8-channel pipette automatic tip pickup."
            )
    elif nozzle_map.physical_nozzle_count == 96:
        if nozzle_map.starting_nozzle == "A1":
            return _get_bottom_to_top_right_to_left(well_list)
        elif nozzle_map.starting_nozzle == "A12":
            return _get_bottom_to_top_left_to_right(well_list)
        elif nozzle_map.starting_nozzle == "H1":
            return _get_top_to_bottom_right_to_left(well_list)
        elif nozzle_map.starting_nozzle == "H12":
            return _get_top_to_bottom_left_to_right(well_list)
        else:
            raise ValueError(
                f"Nozzle {nozzle_map.starting_nozzle} is an invalid starting tip for 96-channel automatic tip pickup."
            )
    else:
        raise ValueError(
            f"Automatic tip pickup does not support {nozzle_map.physical_nozzle_count}-channel pipettes"
        )


def _get_top_to_bottom_left_to_right(well_list: List[List[str]]) -> List[str]:
    return [well for column in well_list for well in column]


def _get_bottom_to_top_left_to_right(well_list: List[List[str]]) -> List[str]:
    reverse_column_ordering = [list(reversed(column)) for column in well_list]
    return [well for column in reverse_column_ordering for well in column]


def _get_top_to_bottom_right_to_left(well_list: List[List[str]]) -> List[str]:
    reverse_row_ordering = list(reversed(well_list))
    return [well for column in reverse_row_ordering for well in column]


def _get_bottom_to_top_right_to_left(well_list: List[List[str]]) -> List[str]:
    reverse_row_column_ordering = [
        list(reversed(column)) for column in reversed(well_list)
    ]
    return [well for column in reverse_row_column_ordering for well in column]



# noqa: D100

import pytest

from opentrons_shared_data.labware import list_definitions as list_labware_definitions

from opentrons.protocol_api.core.engine._default_labware_versions import (
    DEFAULT_LABWARE_VERSIONS,
    KNOWN_EXCEPTIONS_FOR_TESTS,
    get_standard_labware_default_version,
)
from opentrons.protocols.api_support.types import APIVersion

_TEST_DEFAULT_LABWARE_VERSIONS = {
    APIVersion(1, 100): {
        "well_plate_a": 3,
    },
    APIVersion(1, 110): {
        "well_plate_b": 4,
    },
    APIVersion(1, 120): {
        "well_plate_a": 5,
        "well_plate_b": 6,
    },
}


def _get_available_load_names_and_highest_versions() -> list[tuple[str, int]]:
    highest_available_version_by_load_name: dict[str, int] = {}
    for load_name, version, _ in list_labware_definitions():
        if (
            load_name not in highest_available_version_by_load_name
            or version > highest_available_version_by_load_name[load_name]
        ):
            highest_available_version_by_load_name[load_name] = version
    return sorted(highest_available_version_by_load_name.items())


def _highest_possible_default_version(load_name: str) -> int:
    """For a given labware, return the highest version that will ever be loaded by default, across all apiLevels."""
    default_versions_for_this_labware = [
        default_versions_by_load_name[load_name]
        for default_versions_by_load_name in DEFAULT_LABWARE_VERSIONS.values()
        if load_name in default_versions_by_load_name
    ]
    return max(default_versions_for_this_labware, default=1)


@pytest.mark.parametrize(
    ("api_version", "load_name", "expected_labware_version"),
    [
        # At 1.100, well_plate_a upgrades to 3.
        (APIVersion(1, 99), "well_plate_a", 1),
        (APIVersion(1, 99), "well_plate_b", 1),
        (APIVersion(1, 99), "well_plate_c", 1),
        (APIVersion(1, 100), "well_plate_a", 3),
        (APIVersion(1, 100), "well_plate_b", 1),
        (APIVersion(1, 100), "well_plate_c", 1),
        # At 1.110, well_plate_b upgrades to 4.
        (APIVersion(1, 109), "well_plate_a", 3),
        (APIVersion(1, 109), "well_plate_b", 1),
        (APIVersion(1, 109), "well_plate_c", 1),
        (APIVersion(1, 110), "well_plate_a", 3),
        (APIVersion(1, 110), "well_plate_b", 4),
        (APIVersion(1, 110), "well_plate_c", 1),
        # At 1.120, well_plate_a upgrades to 5 and well_plate_b upgrades to 6.
        (APIVersion(1, 119), "well_plate_a", 3),
        (APIVersion(1, 119), "well_plate_b", 4),
        (APIVersion(1, 119), "well_plate_c", 1),
        (APIVersion(1, 120), "well_plate_a", 5),
        (APIVersion(1, 120), "well_plate_b", 6),
        (APIVersion(1, 120), "well_plate_c", 1),
        # Higher versions are the same as the highest-defined version, in this case 1.120.
        (APIVersion(1, 121), "well_plate_a", 5),
        (APIVersion(1, 121), "well_plate_b", 6),
        (APIVersion(1, 121), "well_plate_c", 1),
    ],
)
def test_get_standard_labware_default_version(
    api_version: APIVersion, load_name: str, expected_labware_version: int
) -> None:
    """Test the logic for resolving a single labware's version at an arbitrary api_version."""
    assert (
        get_standard_labware_default_version(
            api_version, load_name, _TEST_DEFAULT_LABWARE_VERSIONS
        )
        == expected_labware_version
    )


@pytest.mark.parametrize(
    ("load_name", "highest_available_version"),
    _get_available_load_names_and_highest_versions(),
)
def test_default_labware_version_coverage(
    load_name: str, highest_available_version: int
) -> None:
    """When new labware versions are added, make sure they're included in some apiLevel."""
    if load_name not in KNOWN_EXCEPTIONS_FOR_TESTS:
        assert (
            _highest_possible_default_version(load_name) == highest_available_version
        ), (
            f"Expected version {highest_available_version} of {load_name} to be included in some apiLevel. See DEFAULT_LABWARE_VERSIONS."
        )
    else:
        assert (
            _highest_possible_default_version(load_name) != highest_available_version
        ), (
            f"{load_name} is included in KNOWN_EXCEPTIONS_FOR_TESTS but it no longer needs to be?"
        )



"""Set tip state command request, result, and implementation models."""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, Type

from pydantic import BaseModel, Field
from typing_extensions import Literal

from ..errors.error_occurrence import ErrorOccurrence
from ..state.update_types import StateUpdate
from ..types import TipRackWellState
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
)

if TYPE_CHECKING:
    from ..state.state import StateView


SetTipStateCommandType = Literal["setTipState"]


class SetTipStateParams(BaseModel):
    """Payload needed to set tip wells of a tip rack to the requested state."""

    labwareId: str = Field(
        ..., description="Identifier of tip rack labware to set tip wells in."
    )
    wellNames: List[str] = Field(
        ..., description="Names of the well to set tip well state for."
    )
    tipWellState: TipRackWellState = Field(
        ..., description="State to set tip wells to."
    )


class SetTipStateResult(BaseModel):
    """Result data from the execution of a setTipState command."""

    pass


class SetTipStateImplementation(
    AbstractCommandImpl[SetTipStateParams, SuccessData[SetTipStateResult]]
):
    """Set tip state command implementation."""

    def __init__(
        self,
        state_view: StateView,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view

    async def execute(
        self, params: SetTipStateParams
    ) -> SuccessData[SetTipStateResult]:
        """Set the tip rack wells to the requested state."""
        labware_id = params.labwareId
        well_names = params.wellNames

        self._state_view.labware.raise_if_not_tip_rack(labware_id=labware_id)
        self._state_view.labware.raise_if_wells_are_invalid(
            labware_id=labware_id, well_names=well_names
        )

        return SuccessData(
            public=SetTipStateResult(),
            state_update=StateUpdate().update_tip_rack_well_state(
                tip_state=params.tipWellState,
                labware_id=labware_id,
                well_names=well_names,
            ),
        )


class SetTipState(BaseCommand[SetTipStateParams, SetTipStateResult, ErrorOccurrence]):
    """Set tip state command model."""

    commandType: SetTipStateCommandType = "setTipState"
    params: SetTipStateParams
    result: Optional[SetTipStateResult] = None

    _ImplementationCls: Type[SetTipStateImplementation] = SetTipStateImplementation


class SetTipStateCreate(BaseCommandCreate[SetTipStateParams]):
    """Set tip state command creation request model."""

    commandType: SetTipStateCommandType = "setTipState"
    params: SetTipStateParams

    _CommandCls: Type[SetTipState] = SetTipState



from typing import List

import pytest

from opentrons.drivers.types import PlateTemperature, Temperature
from opentrons.hardware_control.modules.plate_temp_status import PlateTemperatureStatus
from opentrons.hardware_control.modules.types import TemperatureStatus

TEMP_DIFF_ABOVE_THRESHOLD = PlateTemperatureStatus.TEMP_THRESHOLD * 2

TEMP_DIFF_BELOW_THRESHOLD = PlateTemperatureStatus.TEMP_THRESHOLD / 2

MIN_SAMPLES_FOR_HOLD = PlateTemperatureStatus.MIN_SAMPLES_UNDER_THRESHOLD

ONE_TOO_FEW_SAMPLES = MIN_SAMPLES_FOR_HOLD - 1


@pytest.mark.parametrize(
    argnames=["temps", "expected"],
    argvalues=[
        # No temperature readings
        [[], TemperatureStatus.ERROR],
        # One too few readings
        [
            [Temperature(current=50, target=None)] * ONE_TOO_FEW_SAMPLES,
            TemperatureStatus.IDLE,
        ],
        # No target
        [
            [Temperature(current=50, target=None)] * MIN_SAMPLES_FOR_HOLD,
            TemperatureStatus.IDLE,
        ],
        # All readings at target temperature
        [
            [Temperature(current=50, target=50)] * MIN_SAMPLES_FOR_HOLD,
            TemperatureStatus.HOLDING,
        ],
        # All are close enough. Half the buffer a bit above and half a bit below.
        [
            [
                Temperature(current=50 + TEMP_DIFF_BELOW_THRESHOLD, target=50),
                Temperature(current=50 - TEMP_DIFF_BELOW_THRESHOLD, target=50),
            ]
            * int(MIN_SAMPLES_FOR_HOLD / 2),
            TemperatureStatus.HOLDING,
        ],
        # One reading above target
        [
            [Temperature(current=50 + TEMP_DIFF_ABOVE_THRESHOLD, target=50)],
            TemperatureStatus.COOLING,
        ],
        # One reading below target
        [
            [Temperature(current=50 - TEMP_DIFF_ABOVE_THRESHOLD, target=50)],
            TemperatureStatus.HEATING,
        ],
        # All samples are the same but one that is above threshold
        [
            [Temperature(current=50, target=50)] * ONE_TOO_FEW_SAMPLES
            + [Temperature(current=50 + TEMP_DIFF_ABOVE_THRESHOLD, target=50)],
            TemperatureStatus.COOLING,
        ],
        # All samples are the same but one that is below threshold
        [
            [Temperature(current=50, target=50)] * ONE_TOO_FEW_SAMPLES
            + [Temperature(current=50 - TEMP_DIFF_ABOVE_THRESHOLD, target=50)],
            TemperatureStatus.HEATING,
        ],
    ],
)
def test_plate_temp_status(
    temps: List[PlateTemperature], expected: TemperatureStatus
) -> None:
    """It should set the status correctly based on temperature readings."""
    status = PlateTemperatureStatus()
    for t in temps:
        status.update(t)
    assert status.status == expected



from typing import Callable

import pytest

from opentrons.protocols.models import json_protocol


@pytest.mark.parametrize(
    argnames=["version", "name"],
    argvalues=[
        ["5", "simpleV5"],
        ["4", "simpleV4"],
        ["4", "testModulesProtocol"],
        ["3", "simple"],
        ["3", "testAllAtomicSingleV3"],
    ],
)
def test_json_protocol_model(
    get_json_protocol_fixture: Callable[..., object],
    version: str,
    name: str,
) -> None:
    """It should be parsed and validated correctly."""
    fx = get_json_protocol_fixture(
        fixture_version=version, fixture_name=name, decode=True
    )

    # Create the model
    d = json_protocol.Model.model_validate(fx)

    # Compare the dict created by pydantic to the loaded json
    assert d.model_dump(exclude_unset=True, by_alias=True) == fx



"""Test move-to-coordinates commands."""

from datetime import datetime

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import StallOrCollisionDetectedError

from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.move_to_coordinates import (
    MoveToCoordinatesImplementation,
    MoveToCoordinatesParams,
    MoveToCoordinatesResult,
)
from opentrons.protocol_engine.commands.movement_common import StallOrCollisionError
from opentrons.protocol_engine.execution import MovementHandler
from opentrons.protocol_engine.resources.model_utils import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.types import DeckPoint
from opentrons.types import Point


@pytest.fixture
def subject(
    movement: MovementHandler, model_utils: ModelUtils
) -> MoveToCoordinatesImplementation:
    """Build a command implementation with injected dependencies."""
    return MoveToCoordinatesImplementation(movement=movement, model_utils=model_utils)


async def test_move_to_coordinates_implementation(
    decoy: Decoy, movement: MovementHandler, subject: MoveToCoordinatesImplementation
) -> None:
    """Test the `moveToCoordinates` implementation."""
    params = MoveToCoordinatesParams(
        pipetteId="pipette-id",
        coordinates=DeckPoint(x=1.11, y=2.22, z=3.33),
        minimumZHeight=1234,
        forceDirect=True,
        speed=567.8,
    )

    decoy.when(
        await movement.move_to_coordinates(
            pipette_id="pipette-id",
            deck_coordinates=DeckPoint(x=1.11, y=2.22, z=3.33),
            direct=True,
            additional_min_travel_z=1234,
            speed=567.8,
        )
    ).then_return(Point(x=4.44, y=5.55, z=6.66))

    result = await subject.execute(params=params)

    assert result == SuccessData(
        public=MoveToCoordinatesResult(position=DeckPoint(x=4.44, y=5.55, z=6.66)),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="pipette-id",
                new_location=None,
                new_deck_point=DeckPoint(x=4.44, y=5.55, z=6.66),
            )
        ),
    )


async def test_move_to_coordinates_stall(
    decoy: Decoy,
    movement: MovementHandler,
    model_utils: ModelUtils,
    subject: MoveToCoordinatesImplementation,
) -> None:
    """It should handle stall errors."""
    params = MoveToCoordinatesParams(
        pipetteId="pipette-id",
        coordinates=DeckPoint(x=1.11, y=2.22, z=3.33),
        minimumZHeight=1234,
        forceDirect=True,
        speed=567.8,
    )

    decoy.when(
        await movement.move_to_coordinates(
            pipette_id="pipette-id",
            deck_coordinates=DeckPoint(x=1.11, y=2.22, z=3.33),
            direct=True,
            additional_min_travel_z=1234,
            speed=567.8,
        )
    ).then_raise(StallOrCollisionDetectedError())
    test_id = "test-id"
    timestamp = datetime.now()
    decoy.when(model_utils.get_timestamp()).then_return(timestamp)
    decoy.when(model_utils.generate_id()).then_return(test_id)

    result = await subject.execute(params=params)

    assert result == DefinedErrorData(
        public=StallOrCollisionError.model_construct(
            id=test_id, createdAt=timestamp, wrappedErrors=[matchers.Anything()]
        ),
        state_update=update_types.StateUpdate(pipette_location=update_types.CLEAR),
    )



"""Module Firmware update script."""

import argparse
import asyncio
import os
import re
import subprocess
import sys
from glob import glob
from typing import Dict, Final, List, Optional

from opentrons.drivers.rpi_drivers import usb
from opentrons.hardware_control import modules
from opentrons.hardware_control.execution_manager import ExecutionManager
from opentrons.hardware_control.module_control import MODULE_PORT_REGEX
from opentrons.hardware_control.modules.mod_abc import AbstractModule
from opentrons.hardware_control.modules.update import update_firmware
from opentrons.hardware_control.types import BoardRevision

# Constants for checking if module is back online
ONLINE_RETRIES = 3
DELAY_S = 5


MODULES: Final[Dict[str, str]] = {
    "temp-deck": "tempdeck",
    "mag-deck": "magdeck",
    "thermocycler": "thermocycler",
    "heater-shaker": "heatershaker",
    "absorbance-reader": "absorbancereader",
    "flex-stacker": "flexstacker",
}


def parse_version(filepath: str) -> str:
    """Parse the version string from the filename."""
    _, ext = os.path.splitext(os.path.basename(filepath))
    ext_pattern = re.escape(ext.lstrip("."))
    pattern = rf"@(v\d+(?:\.\d+)*)\.{ext_pattern}"
    match = re.search(pattern, os.path.basename(filepath))
    return match.group(1) if match else ""


def scan_connected_modules() -> List[modules.ModuleAtPort]:
    """Scan for connected modules and return list of
    tuples of serial ports and device names
    """
    discovered_modules = []
    devices = glob("/dev/ot_module*")
    for port in devices:
        symlink_port = port.split("dev/")[1]
        module_at_port = get_module_at_port(symlink_port)
        if module_at_port:
            discovered_modules.append(module_at_port)
    return discovered_modules


def get_module_at_port(port: str) -> Optional[modules.ModuleAtPort]:
    """Given a port, returns either a ModuleAtPort
    if it is a recognized module, or None if not recognized.
    """
    match = MODULE_PORT_REGEX.search(port)
    if match:
        name = match.group(1).lower()
        if name not in modules.MODULE_TYPE_BY_NAME:
            print(f"Unexpected module connected: {name} on {port}")
            return None
        return modules.ModuleAtPort(port=f"/dev/{port}", name=name)
    return None


async def build_module(
    mod: modules.ModuleAtPort, loop: asyncio.AbstractEventLoop
) -> Optional[AbstractModule]:
    try:
        # Get the device path and
        port = subprocess.check_output(["readlink", "-f", mod.port]).decode().strip()
        # remove the symlink for the device, so its freed by the robot-server
        print(f"Removing symlink {mod.port}")
        subprocess.check_call(["unlink", mod.port])
        # wwait some time to let the device teardown
        await asyncio.sleep(2)
        # create an instance of the module using the device path
        return await modules.build(
            port=port,
            usb_port=mod.usb_port,
            type=modules.MODULE_TYPE_BY_NAME[mod.name],
            execution_manager=ExecutionManager(),
            disconnected_callback=lambda *args: None,
            error_callback=lambda *args: None,
            simulating=False,
            hw_control_loop=loop,
        )
    except Exception:
        return None


async def teardown_module(module: AbstractModule) -> None:
    """Tearsdown the module so it can be used again by the robot-server.."""
    name = module.name()
    serial = module.device_info["serial"]
    port_name = module.usb_port.name
    command = f"echo '{port_name}' | tee /sys/bus/usb/drivers/usb"
    print(f"Removing module: {name} {serial}")
    try:
        # stop the poller and disconnect serial
        await module._poller.stop()  # type: ignore
        await module._driver.disconnect()  # type: ignore
        # unbind the device from usb and re-bind to simulate unplug
        subprocess.run(f"{command}/unbind", shell=True, capture_output=True)
        await asyncio.sleep(2)
        subprocess.run(f"{command}/bind", shell=True, capture_output=True)
    except Exception:
        pass


def enable_udev_rules(enable: bool) -> None:
    """Enable/Disable creation of opentrons modules by the hardware controller.

    This is done so the module is not automatically picked up by the server
    while we are updating it.
    """
    rule = "95-opentrons-modules.rules"
    original = f"/etc/udev/rules.d/{rule}"
    destination = f"/var/lib/{rule}"
    src = original if not enable else destination
    dst = destination if not enable else original
    msg = "Disabl" if not enable else "Enabl"
    if not os.path.exists(src):
        sys.exit(f"ERROR: Rule file not found: {src}")

    try:
        print(f"{msg}ing udev: Moving rule {src} -> {dst}")
        subprocess.check_call(["mount", "-o", "remount,rw", "/"])
        subprocess.check_call(["mv", src, dst])
    except Exception as e:
        sys.exit(
            f"ERROR: Could not {msg}e udev rule: {rule}\n{e}",
        )
    finally:
        subprocess.check_call(["mount", "-o", "remount,ro", "/"])
        subprocess.check_call(["udevadm", "control", "--reload-rules"])


def check_dev_exist(port: str) -> bool:
    """True if the device with the given port exists in /dev."""
    try:
        return subprocess.run(["ls", port], capture_output=True).returncode == 0
    except Exception:
        return False


async def main(args: argparse.Namespace) -> None:  # noqa: C901
    """Entry point for script."""
    mod_name = MODULES[args.module]
    target_version = parse_version(args.file)
    if not os.path.exists(args.file):
        sys.exit(f"Invalid filepath: {args.file}")
    if not target_version:
        sys.exit(f"Target version could not be parsed from file: {args.file}")

    print("Setting up...")
    loop = asyncio.get_running_loop()
    usb_bus = usb.USBBus(BoardRevision.FLEX_B2)  # todo: get this from the robot
    print(f"Searching for {mod_name} modules in /dev/")
    mods = scan_connected_modules()
    mods = usb_bus.match_virtual_ports(mods)  # type: ignore
    if not mods:
        print("No modules found")
        return

    # Disable udev rules so modules aren't re-created by the server when they update
    teardown_modules: List[AbstractModule] = []
    enable_udev_rules(False)
    print("\n------------------------------------------")
    for mod in mods:
        if mod_name not in mod.name:
            continue

        # Create an instance of the opentrons module
        print(f"Found mod: {mod.name} at {mod.port}")
        module = await build_module(mod, loop)
        if module is None:
            continue

        name = module.name()
        version = module.device_info["version"]
        serial = module.device_info["serial"]
        model = module.device_info["model"]
        teardown_modules.append(module)
        print(f"Created module: {module.name()} {model} {serial} {version}")

        # Check that the update file is for this module
        file_prefix = module.firmware_prefix()
        if file_prefix not in args.file:
            print(f"ERROR: Target module does not match file: {mod_name} {args.file}")
            continue

        # Check if the module is one we care about
        if args.serial and serial not in args.serial:
            continue

        # Check if the module needs an update
        if version == target_version and not args.force:
            print(f"Module {name} {serial} is up-to-date.")
            continue

        print(f"Updating {name} {model} {serial}: {version} -> {target_version}")
        await update_firmware(module, args.file)

        # wait for the device to come online
        for retry in range(ONLINE_RETRIES):
            if retry >= ONLINE_RETRIES:
                print(f"Module {serial} failed to come back online.")
                break

            print(f"Checking if {name} at {module.port} is online...")
            await asyncio.sleep(DELAY_S)
            if not check_dev_exist(module.port):
                print("Not online")
                continue

            # re-open serial connection
            if not await module._driver.is_connected():  # type: ignore
                await module._driver.connect()  # type: ignore

            # refresh the device info
            print(f"Device {module.port} is back online, refreshing device info.")
            device_info = (await module._driver.get_device_info()).to_dict()  # type: ignore
            success = device_info["version"] == target_version
            msg = "updated successfully!" if success else "failed to update"
            print(f"Device {name} {serial} {msg}")
            break

    print("------------------------------------------\n")
    print("Tearing down")
    # Enable udev rules and teardown the module so they can be pick-up by the robot-server
    enable_udev_rules(True)
    for module in teardown_modules:
        await teardown_module(module)
    print("Done")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Module FW Update.")
    parser.add_argument(
        "--module",
        help="The module target to be updated.",
        type=str,
        required=True,
        choices=MODULES.keys(),
    )
    parser.add_argument(
        "--file",
        help="""Path to binary file containing the FW executable"""
        """Must have format `module-name@vx.x.x.bin/hex`, ex, flex-stacker@v0.0.1.bin""",
        type=str,
        required=True,
    )
    parser.add_argument(
        "--serial",
        help="The specific serial numbers of the devices to update.",
        type=str,
        nargs="+",
    )
    parser.add_argument(
        "--force",
        help="Force install the update, even if the versions are the same.",
        action="store_true",
        default=False,
    )
    args = parser.parse_args()
    try:
        asyncio.run(main(args))
    except Exception as e:
        print("ERROR: Unhandled Exception: ", e)
        # Re-enable udev rules
        enable_udev_rules(True)
